from typing import Callable, Type, Union, Optional

from .module import ModuleMetadata
from .reflect import Reflect


class ClassMetadata:
    Metadata = "__dependency_metadata__"
    _global_providers: list = []
    _module = None

    def __init__(self, module: Callable, global_providers: Optional[list] = None):
        self._module = module
        self._global_providers = global_providers or []

    def get_module(self):
        return self._module

    def get_global_providers(self):
        return self._global_providers

    def get_service_providers(self, module: Union[Type, None] = None):
        providers: list = Reflect.get_metadata(
            module or self._module, ModuleMetadata.Providers, []
        )
        import_providers_form_exports: list = []
        # Only not a root module need to get import_providers to share
        # if not Reflect.get_metadata(self._module, ModuleMetadata.Root, False):
        for im in Reflect.get_metadata(
            module or self._module, ModuleMetadata.Imports, []
        ):
            exports: list = Reflect.get_metadata(
                im.module if hasattr(im, "module") else im, ModuleMetadata.Exports, []
            )
            # check if dynamic module
            # if hasattr(im, 'module') and hasattr(im, 'exports'):
            #     exports = getattr(im, 'exports', [])
            for export in exports:
                # For module re-exporting
                is_module: bool = Reflect.get_metadata(
                    export, ModuleMetadata.Module, False
                )
                if is_module:
                    import_providers_form_exports = (
                        import_providers_form_exports
                        + Reflect.get_metadata(export, ModuleMetadata.Providers, [])
                    )
                else:
                    import_providers_form_exports.append(export)
        return providers, import_providers_form_exports



import inspect
from functools import lru_cache
from typing import (
    ForwardRef,
    Type,
    Union,
    Any,
    Optional,
    Callable,
    Awaitable,
    TYPE_CHECKING,
)

from nestipy.metadata import (
    ClassMetadata,
    CtxDepKey,
    ModuleMetadata,
    ProviderToken,
    Reflect,
)
from .context_container import RequestContextContainer
from .dependency import TypeAnnotated
from .helper import ContainerHelper
from .utils import uniq

if TYPE_CHECKING:
    from .provider import ModuleProviderDict

_INIT = "__init__"


class NestipyContainer:
    """
    Singleton container class responsible for managing service instances and their dependencies.

    This container allows the registration and resolution of services (both transient and singleton),
    manages singleton instances, and handles dependency injection for classes and functions.
    """

    _instance: Union["NestipyContainer", None] = None
    _services: dict = {}
    _global_service_instances: dict = {}
    _singleton_instances: dict = {}
    _singleton_classes: set = set()

    def __new__(cls, *args, **kwargs):
        """
        Ensures the class is a singleton by creating only one instance.
        """
        if cls._instance is None:
            cls._instance = super(NestipyContainer, cls).__new__(cls, *args, **kwargs)
        return cls._instance

    @classmethod
    def get_instance(cls, *args, **kwargs):
        """
        Returns the singleton instance of the container.
        """
        return NestipyContainer(*args, **kwargs)

    def add_transient(self, service: Type):
        """
        Registers a transient service in the container.
        Transient services are created every time they are requested.

        :param service: The service class to be registered.
        """
        self._services[service] = service

    def add_singleton(self, service: Type):
        """
        Registers a singleton service in the container.
        Singleton services are created only once and reused for subsequent requests.

        :param service: The service class to be registered as singleton.
        """
        self._services[service] = service
        self._singleton_classes.add(service)

    def get_all_services(self) -> list:
        """
        Returns a list of all registered services in the container.

        :return: List of service types.
        """
        return list(self._services.keys())

    def add_singleton_instance(
        self, service: Union[Type, str], service_instance: object
    ):
        """
        Registers an instance of a singleton service.

        :param service: The service class or token.
        :param service_instance: The instance of the service to be used.
        """
        self._singleton_instances[service] = service_instance

    def get_all_singleton_instance(self) -> list:
        """
        Returns a list of all singleton instances currently in the container.

        :return: List of singleton instances.
        """
        return [v for k, v in self._singleton_instances.items()]

    @classmethod
    @lru_cache()
    def get_global_providers(cls) -> list:
        """
        Retrieves a list of global providers registered across all modules.

        :return: List of global provider services.
        """
        global_providers = []
        for service in cls._services:
            if service in global_providers:
                continue
            metadata: Union[ClassMetadata | None] = Reflect.get_metadata(
                service, ClassMetadata.Metadata, None
            )
            if metadata is not None:
                is_global = Reflect.get_metadata(
                    metadata.get_module(), ModuleMetadata.Global, False
                )
                is_root = Reflect.get_metadata(
                    metadata.get_module(), ModuleMetadata.Root, False
                )
                if is_global or is_root:
                    global_providers += Reflect.get_metadata(
                        metadata.get_module(), ModuleMetadata.Providers, []
                    )
        return uniq(global_providers)

    @classmethod
    def get_dependency_metadata(cls, service: Union[Type, object]) -> list:
        """
        Retrieves the metadata for the dependencies of a given service.

        :param service: The service class to resolve.
        :return: List of metadata related to the service's dependencies.
        """
        from .provider import ModuleProviderDict

        # extract global data from _service, not from module because all provider is already saved in _services of
        # container
        metadata: Union[ClassMetadata | None] = Reflect.get_metadata(
            service, ClassMetadata.Metadata, None
        )
        if metadata is not None:
            global_providers = cls.get_global_providers()
            providers, import_providers = metadata.get_service_providers()
            uniq_providers = []
            for m in uniq(providers + global_providers + import_providers):
                if isinstance(m, ModuleProviderDict):
                    for imported_module in m.imports:
                        m_provider, m_import_providers = metadata.get_service_providers(
                            imported_module
                        )
                        uniq_providers = uniq(
                            uniq_providers + m_import_providers + m_provider
                        )
                    uniq_providers.append(m.token)
                else:
                    uniq_providers.append(m)
            return uniq(uniq_providers)
        # raise ValueError(f"Dependency Metadata not found  for {service.__name__} service ")
        return []

    @classmethod
    async def _resolve_contextual_service(
        cls, name: str, dep_key: TypeAnnotated, annotation: Union[Type, Any]
    ):
        """
        Resolves a context-specific service based on the provided dependency key.

        :param name: Name of the dependency.
        :param dep_key: Dependency key with metadata.
        :param annotation: Type or annotation for the service.
        :return: The resolved context service.
        """
        context_container = RequestContextContainer.get_instance()
        callback = dep_key.metadata.callback
        if inspect.iscoroutinefunction(callback):
            return await callback(
                name, dep_key.metadata.token, annotation, context_container
            )
        else:
            return callback(name, dep_key.metadata.token, annotation, context_container)

    async def _resolve_module_provider_dict(
        self, instance: "ModuleProviderDict", search_scope: list
    ):
        """
        Resolves a service instance from a ModuleProviderDict.

        :param instance: The ModuleProviderDict instance.
        :param search_scope: The search scope for resolving dependencies.
        :return: The resolved service instance or None.
        """
        if instance.value:
            return instance.value
        elif instance.existing:
            if isinstance(instance.existing, ProviderToken):
                return await self.get(instance.existing.key)
            else:
                return await self.get(instance.existing)
        elif instance.use_class:
            return await self.get(instance.use_class)
        elif instance.factory:
            return await self.resolve_factory(
                factory=instance.factory,
                inject=instance.inject,
                search_scope=search_scope,
            )

        else:
            return None

    async def _check_exist_singleton(self, key: Union[Type, str]):
        """
        Checks if a singleton instance exists for a given service key.

        :param key: The service class or token.
        :return: The singleton instance if it exists, otherwise None.
        """
        from .provider import ModuleProviderDict

        if key in self._singleton_instances:
            instance = self._singleton_instances[key]
            # to keep improve
            if isinstance(instance, ModuleProviderDict):
                search_scope = self.get_dependency_metadata(instance)
                if instance.token in search_scope:
                    value = await self._resolve_module_provider_dict(
                        instance, search_scope=search_scope
                    )
                    # update singleton instance to have the async value from ModuleProviderDict
                    self._singleton_instances[key] = value
                    return value
                else:
                    raise ValueError(
                        f"Service {instance.__class__.__name__} not found in scope"
                    )
            else:
                return instance
        return None

    def _check_service(
        self, key: Union[Type, str], origin: Optional[list] = None
    ) -> tuple:
        """
        Checks if a service is registered and detects circular dependencies.

        :param key: The service class or token.
        :param origin: The list of services currently being resolved to avoid circular dependency.
        :return: The service class and updated origin list.
        :raises ValueError: If the service is not found or circular dependency is detected.
        """
        if key not in self._services:
            raise ValueError(f"Service {key} not found")
        service = self._services[key]
        if service in (origin or []):
            raise ValueError(
                f"Circular dependency found  for {service.__name__} service "
            )
        return service, origin or set()

    async def _resolve_property(
        self,
        key: Union[Type, str],
        origin: Optional[list] = None,
        disable_scope: bool = False,
    ):
        """
        Resolves dependencies for the properties of a service.

        :param key: The service class or token.
        :param origin: The list of services currently being resolved.
        :param disable_scope: Whether to disable scope-based resolution.
        """
        service, origin = self._check_service(key, origin)
        search_scope = self.get_dependency_metadata(service)
        origin.add(service)
        annotations: dict = getattr(service, "__annotations__", {})
        for name, param_annotation in annotations.items():
            annotation, dep_key = ContainerHelper.get_type_from_annotation(
                param_annotation
            )
            if dep_key.metadata.key is not CtxDepKey.Service:
                dependency = await self._resolve_contextual_service(
                    name, dep_key, annotation
                )
                setattr(service, name, dependency)
            else:
                # check token
                if (
                    dep_key.metadata.token in search_scope
                    or annotation in search_scope
                    or disable_scope
                ):
                    key = dep_key.metadata.token or annotation
                    # A modifier
                    if isinstance(key, ForwardRef):
                        key = eval(key.__forward_arg__, globals(), locals())
                    dependency = await self.get(key)
                    setattr(service, name, dependency)
                else:
                    _name: str = (
                        annotation.__name__
                        if not isinstance(annotation, str)
                        else annotation
                    )
                    raise ValueError(
                        f"Service {_name} not found in scope {search_scope}"
                    )
        origin.remove(service)
        self._services[key] = service

    async def _get_method_dependency(
        self,
        method_to_resolve: Callable,
        search_scope: list,
        disable_scope: bool = False,
    ):
        """
        Resolves dependencies for a method's parameters.

        :param method_to_resolve: The method to resolve dependencies for.
        :param search_scope: The search scope for resolving dependencies.
        :param disable_scope: Whether to disable scope-based resolution.
        :return: A dictionary of resolved dependencies for the method.
        """
        params = inspect.signature(method_to_resolve).parameters
        args = {}
        for name, param in params.items():
            if name != "self" and param.annotation is not inspect.Parameter.empty:
                annotation, dep_key = ContainerHelper.get_type_from_annotation(
                    param.annotation
                )
                if dep_key.metadata.key is not CtxDepKey.Service:
                    dependency = await self._resolve_contextual_service(
                        name, dep_key, annotation
                    )
                    args[name] = dependency
                elif (
                    dep_key.metadata.token in search_scope
                    or annotation in search_scope
                    or disable_scope
                ):
                    dependency = await self.get(dep_key.metadata.token or annotation)
                    args[name] = dependency
                else:
                    _name: str = (
                        annotation.__name__
                        if not isinstance(annotation, str)
                        else annotation
                    )
                    raise ValueError(
                        f"Service {_name} not found in scope {search_scope}"
                    )
        return args

    @classmethod
    async def _call_method(cls, method: Callable, args: dict):
        if inspect.iscoroutinefunction(method):
            return await method(**args)
        return method(**args)

    async def resolve_factory(
        self,
        factory: Callable,
        inject: list,
        search_scope: list,
        disable_scope: bool = False,
    ):
        """
        Resolves dependencies and calls a factory function to create an instance.

        :param factory: The factory function to call.
        :param inject: Optional list of dependencies to inject into the factory.
        :param search_scope: The search scope for resolving dependencies.
        :param disable_scope: disable scope for searching dependencies.
        :return: The instance created by the factory.
        """
        search_scope_by_inject = [m for m in inject if m in search_scope]
        args = await self._get_method_dependency(
            method_to_resolve=factory,
            search_scope=search_scope_by_inject,
            disable_scope=disable_scope,
        )
        return await self._call_method(method=factory, args=args)

    async def _resolve_method(
        self,
        key: Union[Type, str, object],
        method: str = _INIT,
        origin: Optional[list] = None,
        disable_scope: bool = False,
    ):
        service, origin = self._check_service(key, origin)
        search_scope = self.get_dependency_metadata(service)
        origin.add(service)
        method_to_resolve = getattr(service, method, None)
        if not method_to_resolve:
            raise Exception(f"Method {method} not found in {service.__name__} service ")
        args = await self._get_method_dependency(
            method_to_resolve, search_scope, disable_scope=disable_scope
        )
        if method == _INIT:
            result = service(**args)
            if service in self._singleton_classes:
                self._singleton_instances[service] = result
        else:
            # Service must be an instance (controller)
            instance = await self.get(key)
            instance_method = getattr(instance, method, method_to_resolve)
            result = await self._call_method(instance_method, args)

        origin.remove(service)
        return result

    async def get(
        self,
        key: Union[Type, str],
        method: str = _INIT,
        origin: Optional[list] = None,
        disable_scope: Optional[bool] = False,
    ) -> Awaitable[object]:
        """
        Retrieves an instance of a service, creating it if necessary.

        :param key: The service class or token to retrieve.
        :param method: The name of the method to resolve.
        :param origin: The list of services currently being resolved.
        :param disable_scope: Disabling scope of search for dependencies.
        :return: The service instance.
        """
        in_singleton = await self._check_exist_singleton(key=key)
        if in_singleton:
            if method == _INIT:
                return in_singleton
        else:
            await self._resolve_property(
                key, origin=origin, disable_scope=disable_scope
            )
        return await self._resolve_method(
            key, method=method, origin=origin, disable_scope=disable_scope
        )



import asyncio

from nestipy.microservice.client import RabbitMQClientProxy, MicroserviceOption
from nestipy.microservice.client.option import RabbitMQClientOption


async def main():
    client = RabbitMQClientProxy(
        MicroserviceOption(option=RabbitMQClientOption(host="localhost", port=5672)),
    )
    await client.send("topic", "request")


if __name__ == "__main__":
    asyncio.run(main())



import dataclasses
import inspect
from abc import ABC, abstractmethod
from base64 import b64decode, b64encode
from typing import Type, Union, Literal, Any, Callable
from uuid import uuid4

import itsdangerous
import ujson
from itsdangerous.exc import BadSignature

from nestipy.common.decorator import Injectable
from nestipy.common.http_ import Request, Response
from nestipy.common.middleware.interface import NestipyMiddleware
from nestipy.types_ import NextFn


class SessionStore(ABC):
    def __init__(self):
        self.store = {}

    @abstractmethod
    def get(self, session_id: str) -> dict:
        raise NotImplemented("Must implement")

    @abstractmethod
    def set(self, session_id: str, session_data: Any) -> None:
        raise NotImplemented("Must implement")

    @abstractmethod
    def delete(self, session_id: str) -> None:
        raise NotImplemented("Must implement")


class InMemoryStore(SessionStore):
    def __init__(self):
        super().__init__()

    def get(self, session_id: str) -> dict:
        return self.store.get(session_id, {})

    def set(self, session_id: str, session_data: Any) -> None:
        self.store[session_id] = session_data

    def delete(self, session_id: str) -> None:
        if session_id in self.store:
            del self.store[session_id]


@dataclasses.dataclass
class SessionCookieOption:
    secret_key: str = ""
    session_cookie: str = "cookie-session"
    max_age: Union[int, None] = 60 * 60
    path: str = "/"
    same_site: Literal["lax", "strict", "none"] = "lax"
    https_only: bool = False
    domain: Union[str, None] = None


def cookie_session(option: SessionCookieOption = SessionCookieOption()) -> Type:
    @Injectable()
    class CookieSessionMiddleware(NestipyMiddleware):
        async def use(self, req: Request, res: Response, next_fn: NextFn):
            initial_session_was_empty = True
            signer = itsdangerous.TimestampSigner(str(option.secret_key))
            security_flags = "httponly; samesite=" + option.same_site
            if option.https_only:
                security_flags += "; secure"
            if option.domain is not None:
                security_flags += f"; domain={option.domain}"
            if option.session_cookie in req.cookies:
                data = req.cookies[option.session_cookie].encode("utf-8")
                try:
                    un_sign_data = signer.unsign(data, max_age=option.max_age)
                    req.session = {
                        **req.session,
                        **ujson.loads(b64decode(un_sign_data)),
                    }
                    initial_session_was_empty = False
                except BadSignature:
                    pass
            try:
                result = await next_fn()
            finally:
                cookie_max_age = f"Max-Age={option.max_age}; " if option.max_age else ""
                header_value = ", ".join(
                    [
                        f"{key}={value}; path={option.path}; {cookie_max_age}{security_flags}"
                        for (key, value) in res.cookies()
                    ]
                )
                if header_value:
                    header_value += ", "
                if req.session:
                    data = b64encode(ujson.dumps(req.session).encode("utf-8"))
                    sign_data = signer.sign(data)
                    header_value += "{session_cookie}={data}; path={path}; {max_age}{security_flags}".format(
                        # noqa E501
                        session_cookie=option.session_cookie,
                        data=sign_data.decode("utf-8"),
                        path=option.path,
                        max_age=f"Max-Age={option.max_age}; " if option.max_age else "",
                        security_flags=security_flags,
                    )
                    res.header("Set-Cookie", header_value)
                elif initial_session_was_empty:
                    header_value += "{session_cookie}={data}; path={path}; {expires}{security_flags}".format(
                        # noqa E501
                        session_cookie=option.session_cookie,
                        data="null",
                        path=option.path,
                        expires="expires=Thu, 01 Jan 1970 00:00:00 GMT; ",
                        security_flags=security_flags,
                    )
                    res.header("Set-Cookie", header_value)

            return result

    return CookieSessionMiddleware


@dataclasses.dataclass
class SessionOption:
    secret_key: str = ""
    session_key: str = "session"
    max_age: Union[int, None] = 60 * 60
    path: str = "/"
    same_site: Literal["lax", "strict", "none"] = "lax"
    https_only: bool = False
    domain: Union[str, None] = None
    store: SessionStore = InMemoryStore()


async def call(callback: Callable, *args, **kwargs):
    if inspect.iscoroutinefunction(callback):
        return await callback(*args, **kwargs)
    else:
        return callback(*args, **kwargs)


def session(option: SessionOption = SessionOption()):
    @Injectable()
    class SessionMiddleware(NestipyMiddleware):
        async def use(self, req: "Request", res: "Response", next_fn: "NextFn"):
            initial_session_was_empty = True
            security_flags = "httponly; samesite=" + option.same_site
            if option.https_only:
                security_flags += "; secure"
            if option.domain is not None:
                security_flags += f"; domain={option.domain}"

            session_id = req.cookies.get(option.session_key)

            if session_id and session_id in option.store.store:
                sessions = await call(option.store.get, session_id)
                req.session = {**req.session, **sessions}
                initial_session_was_empty = False
            else:
                session_id = str(uuid4())
            try:
                result = await next_fn()
            finally:
                cookie_max_age = f"Max-Age={option.max_age}; " if option.max_age else ""
                header_value = ", ".join(
                    [
                        f"{key}={value}; path={option.path}; {cookie_max_age}{security_flags}"
                        for (key, value) in res.cookies()
                    ]
                )
                if header_value:
                    header_value += ", "

                await call(option.store.set, session_id, req.session)
                if req.session:
                    header_value += "{session_key}={data}; path={path}; {max_age}{security_flags}".format(
                        session_key=option.session_key,
                        data=session_id,
                        path=option.path,
                        max_age=f"Max-Age={option.max_age}; " if option.max_age else "",
                        security_flags=security_flags,
                    )
                    res.header("Set-Cookie", header_value)
                elif initial_session_was_empty:
                    header_value += "{session_key}={data}; path={path}; {expires}{security_flags}".format(
                        session_key=option.session_key,
                        data="null",
                        path=option.path,
                        expires="expires=Thu, 01 Jan 1970 00:00:00 GMT; ",
                        security_flags=security_flags,
                    )
                    res.header("Set-Cookie", header_value)

            return result

    return SessionMiddleware



import enum
import inspect
import typing
from typing import Type, Any

from nestipy.dynamic_module import DynamicModule
from nestipy.dynamic_module.module.consumer import MiddlewareConsumer
from nestipy.dynamic_module.module.interface import NestipyModule
from nestipy.ioc import NestipyContainer, ModuleProviderDict
from nestipy.metadata import ModuleMetadata, Reflect
from nestipy.core.providers.discover import DiscoverService
from .on_destroy import OnDestroy
from .on_init import OnInit
from ..graphql.graphql_module import GraphqlModule


class InstanceType(enum.Enum):
    providers: str = "Provider"
    controller: str = ("Controller",)
    module: str = "Module"


class InstanceLoader:
    """
    Create all instance of controller and providers, resolvers, middleware
    use before route mapper
    """

    _is_controller: bool = False
    _module_instances: list = []
    graphql_instance: GraphqlModule = None
    discover: DiscoverService = DiscoverService()

    async def create_instances(self, modules: list[Type]) -> GraphqlModule:
        from nestipy.common import (
            NestipyInterceptor,
            CanActivate,
            NestipyMiddleware,
            ExceptionFilter,
        )

        self.discover: DiscoverService = await NestipyContainer.get_instance().get(
            DiscoverService
        )
        for module in modules:
            if isinstance(module, DynamicModule):
                module = module.module
            if module in self._module_instances:
                continue
            providers = await self._create_providers(module)
            self.discover.add_provider(*providers)
            self._is_controller = True
            controllers = await self._create_controllers(module)
            self.discover.add_controller(*controllers)
            self._is_controller = False
            instance = await self.create_instance(module)
            self._module_instances.append(module)
            self.discover.add_module(instance)
            if isinstance(instance, GraphqlModule):
                self.graphql_instance = instance
            imports = Reflect.get_metadata(module, ModuleMetadata.Imports, [])
            await self.create_instances(imports)
        # Create  NestipyInterceptor, CanActivate, NestipyMiddleware,ExceptionFilter without scope
        container = NestipyContainer.get_instance()
        all_services = container.get_all_services()
        for service in [
            s
            for s in all_services
            if inspect.isclass(s)
            and issubclass(
                s, (NestipyInterceptor, CanActivate, NestipyMiddleware, ExceptionFilter)
            )
        ]:
            await self.create_instance(service, with_scope=False)
        return self.graphql_instance

    async def _create_providers(self, module: Type) -> list:
        provider_instance: list = []
        for provider in Reflect.get_metadata(module, ModuleMetadata.Providers, []):
            if isinstance(provider, ModuleProviderDict):
                continue
            ins = await self.create_instance(provider)
            provider_instance.append(ins)
        return provider_instance

    async def _create_controllers(self, module: Type) -> list:
        controller_instance = []
        for controller in Reflect.get_metadata(module, ModuleMetadata.Controllers, []):
            ins = await self.create_instance(controller)
            controller_instance.append(ins)
        return controller_instance

    async def create_instance(self, class_ref: Type, with_scope: bool = True) -> object:
        instance = await NestipyContainer.get_instance().get(
            class_ref, disable_scope=not with_scope
        )
        await self.initialize_instance(class_ref, instance)
        return instance

    @classmethod
    async def initialize_instance(cls, class_ref: Type, instance: Any) -> None:
        # call configure for module
        if issubclass(class_ref, NestipyModule):
            consumer = MiddlewareConsumer(module=class_ref)
            module = typing.cast(NestipyModule, instance)
            module.configure(consumer)
            await module.on_startup()
        elif issubclass(class_ref, OnInit):
            ins = typing.cast(OnInit, instance)
            await ins.on_startup()

    async def destroy(self):
        for provider in self.discover.get_all_provider():
            if isinstance(provider, OnDestroy):
                await provider.on_shutdown()

        for controller in self.discover.get_all_controller():
            if isinstance(controller, OnDestroy):
                await controller.on_shutdown()

        for module in self.discover.get_all_module():
            if isinstance(module, NestipyModule):
                await module.on_shutdown()



from typing import Callable

from .request import Request


class Websocket(Request):
    def __init__(self, scope: dict, receive: Callable, send: Callable):
        assert scope["type"] == "websocket"
        self.subprotocols = scope.get("subprotocols")
        super().__init__(scope, receive, send)



import asyncio
from typing import Optional
from urllib.parse import urlparse, parse_qs

from nestipy.core.nestipy_application import NestipyApplication
from .helpers import (
    HeadersType,
    QueryType,
    CookiesType,
    get_example_scope,
    TestResponse,
)


class TestSimulator:
    __test__ = False

    def __init__(self, app: NestipyApplication):
        self.app = app
        self.scope: dict = {}
        self.receive_queue = asyncio.Queue()
        self.send_queue = asyncio.Queue()

    @classmethod
    def _create_scope(
        cls,
        method: str,
        path: str,
        headers: Optional[HeadersType] = None,
        query: Optional[QueryType] = None,
        cookies: Optional[CookiesType] = None,
    ) -> dict:
        """Creates a mocked ASGI scope"""
        return get_example_scope(
            method, path, extra_headers=headers, query=query, cookies=cookies
        )

    async def _simulate_receive(self):
        return await self.receive_queue.get()

    async def _simulate_send(self, message: dict):
        await self.send_queue.put(message)

    @classmethod
    def _extract_query_params(cls, path) -> QueryType:
        parsed_url = urlparse(path)
        return parse_qs(parsed_url.query)

    async def make_request(
        self, method: str, path: str, headers: Optional[dict[str, str]] = None, body=b""
    ):
        if headers is not None:
            headers = [
                (key.encode("utf-8"), value.encode("utf-8")) for key, value in headers
            ]
        # app request
        self.scope = self._create_scope(
            method, urlparse(path).path, headers, query=self._extract_query_params(path)
        )
        await self.receive_queue.put(
            {
                "type": "http.request",
                "body": body,
                "more_body": False,
            }
        )
        await self.app.setup()
        await self.app(self.scope, self._simulate_receive, self._simulate_send)
        response = []
        while not self.send_queue.empty():
            response.append(await self.send_queue.get())
        return TestResponse.from_dict(response)



from nestipy.ioc import ModuleProviderDict
from .config import ConfigModule, ConfigService
from .decorator import (
    Module,
    Controller,
    Injectable,
    Route,
    Get,
    Post,
    Patch,
    Put,
    Delete,
    apply_decorators,
)
from .exception import (
    HttpException,
    HttpStatusMessages,
    HttpStatus,
    UseFilters,
    Catch,
    ExceptionKey,
    ExceptionFilter,
)
from .guards import CanActivate, UseGuards
from .http_ import Request, Response, Websocket, UploadFile
from .interceptor import NestipyInterceptor, UseInterceptors
from .logger import logger, console
from .middleware import cors, NestipyMiddleware, session, SessionOption, helmet
from .middleware import cookie_session, SessionCookieOption, SessionStore
from .template import Render, TemplateEngine

__all__ = [
    "ConfigModule",
    "ConfigService",
    "Module",
    "Controller",
    "Injectable",
    "Route",
    "Get",
    "Post",
    "Put",
    "Patch",
    "Delete",
    "HttpException",
    "HttpStatusMessages",
    "HttpStatus",
    "UseFilters",
    "Catch",
    "ExceptionKey",
    "ExceptionFilter",
    "CanActivate",
    "UseGuards",
    "Response",
    "Request",
    "Websocket",
    "UploadFile",
    "NestipyInterceptor",
    "UseInterceptors",
    "cors",
    "helmet",
    "NestipyMiddleware",
    "Render",
    "TemplateEngine",
    "ModuleProviderDict",
    "session",
    "SessionOption",
    "logger",
    "console",
    "apply_decorators",
    "cookie_session",
    "SessionStore",
    "SessionCookieOption",
]



from typing import Optional, Type

from .commander import NestipyCommander


class CommandFactory:
    @classmethod
    def create(cls, root_module: Optional[Type]) -> NestipyCommander:
        instance = NestipyCommander()
        instance.init(root_module)
        return instance



import json
from dataclasses import dataclass
from typing import List, Tuple, Dict, Any
from typing import Sequence, Union
from urllib.parse import quote, urlencode

_DEFAULT_AGENT = (
    b"Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:63.0) Gecko/20100101 Firefox/63.0"
)

_DEFAULT_ACCEPT = b"text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8"

_DEFAULT_ACCEPT_LANGUAGE = b"en-US,en;q=0.9,it-IT;q=0.8,it;q=0.7"

_DEFAULT_ACCEPT_ENCODING = b"gzip, deflate"


def _get_tuple(value: Union[List, Tuple[str, int]]) -> tuple:
    if isinstance(value, tuple):
        return value
    assert len(value) == 2
    return tuple(value)  # type: ignore


HeadersType = Union[None, Sequence[Tuple[bytes, bytes]], Dict[str, str]]
CookiesType = Union[None, Sequence[Tuple[bytes, bytes]], Dict[str, str]]
QueryType = Union[None, bytes, str, dict, list]


def get_example_scope(
    method: str,
    path: str,
    extra_headers: HeadersType = None,
    *,
    query: QueryType = None,
    scheme: str = "http",
    server: Union[None, List, Tuple[str, int]] = None,
    client: Union[None, List, Tuple[str, int]] = None,
    user_agent: bytes = _DEFAULT_AGENT,
    accept: bytes = _DEFAULT_ACCEPT,
    accept_language: bytes = _DEFAULT_ACCEPT_LANGUAGE,
    accept_encoding: bytes = _DEFAULT_ACCEPT_ENCODING,
    cookies: CookiesType = None,
):
    """Returns a mocked ASGI scope"""
    if "?" in path:
        raise ValueError(
            "The path in ASGI messages does not contain query string, "
            "use the `query` parameter"
        )

    if server is None:
        server = ("127.0.0.1", 8000)
    else:
        server = _get_tuple(server)

    if client is None:
        client = ("127.0.0.1", 51492)
    else:
        client = _get_tuple(client)

    server_port = server[1]
    if scheme == "http" and server_port == 80:
        port_part = ""
    elif scheme == "https" and server_port == 443:
        port_part = ""
    else:
        port_part = f":{server_port}"

    host = f"{server[0]}{port_part}"

    if isinstance(extra_headers, dict):
        extra_headers = [
            (key.encode(), value.encode()) for key, value in extra_headers.items()
        ]

    query_string: bytes = b""
    if query:
        if isinstance(query, list):
            query = dict(query)
        if isinstance(query, dict):
            query_string = urlencode(query).encode()
        if isinstance(query, str):
            query_string = query.encode()
        if isinstance(query, bytes):
            query_string = query

    cookies_headers: List[Tuple[bytes, bytes]] = []

    if cookies:
        if isinstance(cookies, list):
            cookies_headers = [
                (b"cookie", quote(key).encode() + b"=" + quote(value).encode())
                for key, value in cookies
            ]
        elif isinstance(cookies, dict):
            cookies_headers = [
                (b"cookie", quote(key).encode() + b"=" + quote(value).encode())
                for key, value in cookies.items()
            ]

    headers = (
        [
            (b"host", host.encode()),
            (b"user-agent", user_agent),
            (b"accept", accept),
            (b"accept-language", accept_language),
            (b"accept-encoding", accept_encoding),
            (b"connection", b"keep-alive"),
            (b"upgrade-insecure-requests", b"1"),
        ]
        + ([tuple(header) for header in extra_headers] if extra_headers else [])
        + cookies_headers
    )

    return {
        "type": scheme,
        "http_version": "1.1",
        "server": tuple(server),
        "client": client,
        "scheme": scheme,
        "method": method,
        "root_path": "",
        "path": path,
        "raw_path": path.encode(),
        "query_string": query_string,
        "headers": headers,
    }


@dataclass
class HTTPBody:
    type: str
    body: bytes


@dataclass
class HTTPResponse:
    type: str
    status: int
    headers: List[Tuple[bytes, bytes]]
    body: HTTPBody


@dataclass
class TestResponse:
    response: HTTPResponse
    raw_response: List[Dict[str, Any]]

    @classmethod
    def from_dict(cls, response: List[Dict[str, Any]]) -> "TestResponse":
        response_body = HTTPBody(type=response[1]["type"], body=response[1]["body"])
        http_response = HTTPResponse(
            type=response[0]["type"],
            status=response[0]["status"],
            headers=response[0]["headers"],
            body=response_body,
        )
        return cls(raw_response=response, response=http_response)

    def status(self) -> int:
        return self.response.status

    def get_headers(self) -> Dict[str, str]:
        headers = {}
        for key, value in self.response.headers:
            headers[key.decode("utf-8")] = value.decode("utf-8")
        return headers

    def body(self) -> bytes:
        return self.response.body.body

    def json(self) -> Any:
        try:
            return json.loads(self.body())
        except json.JSONDecodeError:
            return None



class NestipyGraphqlKey:
    resolver: str = "__graphql_resolver__"
    field_resolver: str = "__field_graphql_resolver__"
    query: str = "__graphql_query__"
    mutation: str = "__graphql_mutation__"
    subscription: str = "__graphql_subscription__"
    handler: str = "__graphql_handler__"
    return_type: str = "__graphql_return_type__"
    args: str = "__graphql_args__"
    kwargs: str = "__graphql_kwargs__"



import inspect
import traceback
from typing import Type, Union, Callable, Any, TYPE_CHECKING

from nestipy.common.logger import logger
from nestipy.ioc import NestipyContainer
from nestipy.ioc import RequestContextContainer
from nestipy.metadata import Reflect, ModuleMetadata
from .decorator import GATEWAY_KEY, EVENT_KEY, SUCCESS_EVENT_KEY, ERROR_EVENT_KEY
from ..core.context.execution_context import ExecutionContext

if TYPE_CHECKING:
    from ..core.adapter.http_adapter import HttpAdapter


class IoSocketProxy:
    def __init__(self, adapter: "HttpAdapter"):
        self.adapter = adapter

    def apply_routes(self, modules: list[Union[Type, object]]):
        io_adapter = self.adapter.get_io_adapter()
        for module_ref in modules:
            gateways = self._get_module_provider_gateway(module_ref)
            for gateway in gateways:
                subscribers = self._get_method_event_subscriber(gateway)
                namespace = Reflect.get_metadata(gateway, GATEWAY_KEY, "/")
                for method_name in subscribers:
                    gateway_method = getattr(gateway, method_name)
                    event_name = Reflect.get_metadata(gateway_method, EVENT_KEY, "*")
                    handler = self._create_io_handler(
                        module_ref, gateway, method_name, namespace, event_name
                    )
                    io_adapter.on(event=event_name, namespace=namespace)(handler)

    @classmethod
    def _get_module_provider_gateway(
        cls, module_ref: Union[Type, object]
    ) -> list[Type]:
        providers = Reflect.get_metadata(module_ref, ModuleMetadata.Providers, [])
        return [
            p
            for p in providers
            if Reflect.get_metadata(p, GATEWAY_KEY, None) is not None
        ]

    @classmethod
    def _get_method_event_subscriber(cls, gateway: Type) -> list[str]:
        methods = []
        for method_name, _ in inspect.getmembers(gateway, inspect.isfunction):
            if method_name.startswith("__"):
                continue
            if (
                Reflect.get_metadata(getattr(gateway, method_name), EVENT_KEY, None)
                is not None
            ):
                methods.append(method_name)
        return methods

    @classmethod
    async def call_handler(cls, handler: Callable, params: dict):
        if inspect.iscoroutinefunction(handler):
            return await handler(**params)
        else:
            return handler(**params)

    def _create_io_handler(
        self,
        module_ref: Type,
        gateway: Type,
        method_name: str,
        namespace: str,
        event_name: Any,
    ) -> Callable[..., Any]:
        async def io_handler(event: Any, client: Any, data: Any):
            io_adapter = self.adapter.get_io_adapter()
            context_container = RequestContextContainer.get_instance()
            container = NestipyContainer.get_instance()
            gateway_method = getattr(gateway, method_name)
            execution_context = ExecutionContext(
                self.adapter,
                module_ref,
                gateway,
                gateway_method,
                None,
                None,
                None,
                None,
                io_adapter,
                client,
                data,
            )
            context_container.set_execution_context(execution_context)
            try:
                result = await container.get(gateway, method_name)
                if result is not None:
                    # get success event from handler
                    success_event = Reflect.get_metadata(
                        gateway, SUCCESS_EVENT_KEY, None
                    )
                    # send response to websocket
                    await self.call_handler(
                        io_adapter.emit,
                        {
                            "event": success_event or event_name,
                            "data": result,
                            "namespace": namespace,
                            "to": client.sid,
                        },
                    )
            except Exception as e:
                tb = traceback.format_exc()
                logger.error(e)
                logger.error(tb)
                error_event = Reflect.get_metadata(gateway, ERROR_EVENT_KEY, None)
                if error_event is not None:
                    await self.call_handler(
                        io_adapter.emit,
                        {
                            "event": error_event,
                            "data": str(e),
                            "namespace": namespace,
                            "to": client.sid,
                        },
                    )
                # create exception handler
            finally:
                context_container.destroy()

        return io_handler



from typing import Optional, Type

from nestipy.ioc.dependency import create_type_annotated
from nestipy.ioc.context_container import RequestContextContainer
from nestipy.metadata import CtxDepKey


def args_callback(
    _name: str,
    _token: Optional[str],
    _type_ref: Type,
    _request_context: RequestContextContainer,
):
    args: dict = _request_context.execution_context.switch_to_graphql().get_args()
    return args.get(_token or _name)


def Root():
    return create_type_annotated(args_callback, CtxDepKey.Args)("root")


def Info():
    return create_type_annotated(args_callback, CtxDepKey.Args)("info")



from typing import Type

from nestipy.common.decorator import Injectable

from nestipy.types_ import NextFn
from nestipy.common.middleware.interface import NestipyMiddleware
from nestipy.common.http_ import Request, Response


def helmet() -> Type:
    """
    Returns:
        HelmetMiddleware(Callable): Helmet middlewares for Nestipy
    """

    @Injectable()
    class HelmetMiddleware(NestipyMiddleware):
        async def use(self, req: Request, res: Response, next_fn: NextFn):
            try:
                result = await next_fn()
            finally:
                res.header("X-Content-Type-Options", "nosniff")
                res.header("X-Frame-Options", "DENY")
                res.header("X-XSS-Protection", "1; mode=block")
                res.header(
                    "Strict-Transport-Security", "max-age=31536000; includeSubDomains"
                )
                res.header("Content-Security-Policy", "default-src 'self'")
                res.header("Referrer-Policy", "no-referrer")
            return result

    return HelmetMiddleware



from strawberry import directive as Directive, schema_directive as SchemaDirective
from strawberry import (
    field as Field,
    lazy as Lazy,
    parent as Parent,
    argument as Argument,
)
from strawberry import input as Input, scalar as Scalar, enum as Enum
from strawberry import type as ObjectType, interface as Interface

from .strawberry_asgi import StrawberryASGI
from .strawberry_adapter import StrawberryAdapter
from .dependency import Root, Info

__all__ = [
    "StrawberryASGI",
    "StrawberryAdapter",
    "ObjectType",
    "Interface",
    "Input",
    "Scalar",
    "Directive",
    "SchemaDirective",
    "Enum",
    "Field",
    "Lazy",
    "Lazy",
    "Parent",
    "Argument",
    "Root",
    "Info",
]



from typing import Optional

from nestipy.core.nestipy_application import NestipyApplication
from .helpers import TestResponse
from .simulator import TestSimulator


class TestClient:
    __test__ = False

    def __init__(self, app: NestipyApplication):
        self._simulator = TestSimulator(app)

    async def get(
        self, path: str, headers: Optional[dict[str, str]] = None, body: bytes = b""
    ) -> TestResponse:
        return await self._simulator.make_request("GET", path, headers, body)

    async def post(
        self, path: str, headers: Optional[dict[str, str]] = None, body: bytes = b""
    ) -> TestResponse:
        return await self._simulator.make_request("POST", path, headers, body)

    async def put(
        self, path: str, headers: Optional[dict[str, str]] = None, body: bytes = b""
    ) -> TestResponse:
        return await self._simulator.make_request("PUT", path, headers, body)

    async def delete(
        self, path: str, headers: Optional[dict[str, str]] = None, body: bytes = b""
    ) -> TestResponse:
        return await self._simulator.make_request("DELETE", path, headers, body)

    async def options(
        self, path: str, headers: Optional[dict[str, str]] = None, body: bytes = b""
    ) -> TestResponse:
        return await self._simulator.make_request("OPTIONS", path, headers, body)



from typing import Type

from nestipy.common.decorator import Injectable

from nestipy.types_ import NextFn
from nestipy.common.middleware.interface import NestipyMiddleware
from nestipy.common.http_ import Request, Response


def cors() -> Type:
    """
    Returns:
        CorsMiddleware(Callable): Cors middlewares for Nestipy
    """

    @Injectable()
    class CorsMiddleware(NestipyMiddleware):
        async def use(self, req: Request, res: Response, next_fn: NextFn):
            try:
                result = await next_fn()
            finally:
                res.header("access-control-allow-origin", "*")
                res.header("access-control-allow-headers", "content-type")
                res.header(
                    "access-control-allow-methods", "GET, POST, PUT, DELETE, OPTIONS"
                )
            return result

    return CorsMiddleware



import asyncio
import typing
from typing import AsyncIterator, Optional

from rich.style import Style

from .base import ClientProxy, MicroserviceOption
from .option import TCPClientOption

__SPLIT__ = "__SPLIT__"

from nestipy.common.logger import console


class TCPServer:
    def __init__(self, host: str = "0.0.0.0", port: int = 1333, verbose: bool = True):
        self.host = host
        self.port = port
        self.verbose = verbose
        self.subscribers: dict[str, list[asyncio.StreamWriter]] = {}

    async def handle_client(
        self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter
    ):
        try:
            while True:
                # Read data from client
                data = await reader.readline()
                if not data:
                    break

                command, topic, *message = (
                    data.decode("utf-8").strip().split(__SPLIT__, 2)
                )
                message = message[0] if message else ""
                if command == "SUBSCRIBE":
                    await self.subscribe(topic, writer)
                elif command == "PUBLISH":
                    await self.publish(topic, message)

        except Exception as e:
            if self.verbose:
                console.print(
                    f"ERROR:     {e}",
                    style=Style(color="red", bold=True),
                )
        finally:
            writer.close()
            try:
                await writer.wait_closed()
            finally:
                pass

    async def subscribe(self, topic: str, writer: asyncio.StreamWriter):
        """Add a subscriber to a topic."""
        if topic not in self.subscribers:
            self.subscribers[topic] = []
        self.subscribers[topic].append(writer)
        if self.verbose:
            console.print(
                f"INFO:     Client subscribed to {topic}",
                style=Style(color="green", bold=True),
            )

    async def publish(self, topic: str, message: str):
        """Broadcast a message to all subscribers of a topic."""
        if topic in self.subscribers:
            for sub in self.subscribers[topic]:
                try:
                    sub.write(f"[{topic}]{__SPLIT__}{message}\n".encode("utf-8"))
                    await sub.drain()
                except Exception as e:
                    self.subscribers[topic].remove(sub)
        if self.verbose:
            console.print(
                f"INFO:     Published to {topic}: {message}",
                style=Style(color="green", bold=True),
            )

    async def start(self):
        server = await asyncio.start_server(self.handle_client, self.host, self.port)
        addr = server.sockets[0].getsockname()
        if self.verbose:
            console.print(
                f"INFO:     Server started on {':'.join(addr)}...",
                style=Style(color="green", bold=True),
            )
        async with server:
            await server.serve_forever()


class TCPClientProxy(ClientProxy):
    def __init__(self, option: MicroserviceOption):
        super().__init__(option)
        self.reader: Optional[asyncio.StreamReader] = None
        self.writer: Optional[asyncio.StreamWriter] = None
        self._response_cache = {}
        self._config = typing.cast(
            TCPClientOption, self.option.option or TCPClientOption()
        )
        self.server: Optional[TCPServer] = (
            TCPServer(
                host=self._config.host,
                port=self._config.port,
                verbose=self._config.verbose,
            )
            if self._config.start_server
            else None
        )
        self.server_task: Optional[asyncio.Task] = None

    async def before_start(self):
        """Run before starting the microservice server listener"""
        if self.server is not None:
            self.server_task = asyncio.get_running_loop().create_task(
                self.server.start(), name="tcp_server"
            )
            await asyncio.sleep(1)

    async def before_close(self):
        """Run before stopping the microservice server listener"""
        if self.server_task:
            self.server_task.cancel()

    async def connect(self):
        """Establish a TCP connection."""
        if not self.writer:
            self.reader, self.writer = await asyncio.open_connection(
                self._config.host, self._config.port
            )

    async def slave(self) -> "TCPClientProxy":
        """Return a new instance for slave connection."""
        return TCPClientProxy(self.option)

    async def _publish(self, topic: str, data: str):
        """Publish a message to the specified topic."""
        await self.connect()
        self.writer.write(
            f"PUBLISH{__SPLIT__}{topic}{__SPLIT__}{data}\n".encode("utf-8")
        )
        await self.writer.drain()

    async def subscribe(self, topic: str):
        """Subscribe to a topic."""
        await self.connect()
        self.writer.write(f"SUBSCRIBE{__SPLIT__}{topic}\n".encode("utf-8"))
        await self.writer.drain()

    async def unsubscribe(self, topic: str):
        """Unsubscribe from a topic."""
        await self.connect()
        self.writer.write(f"UNSUBSCRIBE{__SPLIT__}{topic}\n".encode("utf-8"))
        await self.writer.drain()

    async def send_response(self, topic: str, data: str):
        """Send a response to a topic."""
        await self._publish(topic, data)

    async def listen(self) -> AsyncIterator[str]:
        """Listen for incoming messages."""
        while True:
            data = await self.reader.readline()
            if data:
                yield data.decode("utf-8").strip().split(__SPLIT__)[-1]

    async def listen_response(self, from_topic: str, timeout: int = 30) -> str:
        """Listen for a response on a given topic."""
        await self.subscribe(from_topic)
        try:
            while True:
                data = await self.reader.readline()
                if data:
                    message = data.decode("utf-8").strip()
                    key = f"[{from_topic}]{__SPLIT__}"
                    if message.startswith(key):
                        return message[len(key) :]
                await asyncio.sleep(0.01)
        finally:
            await self.unsubscribe(from_topic)

    async def close(self):
        """Close the connection."""
        if self.writer:
            self.writer.close()
            await self.writer.wait_closed()
            self.writer = None



import traceback
from typing import Type, Optional, cast

from .style import echo

from nestipy.common.utils import uniq_list
from nestipy.core.instance_loader import InstanceLoader
from nestipy.core.meta.controller_metadata_creator import ControllerMetadataCreator
from nestipy.core.meta.module_metadata_creator import ModuleMetadataCreator
from nestipy.core.meta.provider_metadata_creator import ProviderMetadataCreator
from nestipy.dynamic_module import DynamicModule
from nestipy.metadata import ModuleMetadata
from nestipy.metadata.reflect import Reflect
from .abstract import BaseCommand
from .meta import CommanderMeta


class NestipyCommander(object):
    _root_module: Optional[Type] = None

    def __init__(self):
        self.instance_loader = InstanceLoader()

    @classmethod
    def _get_modules(cls, module: Type) -> list[Type]:
        modules: list[Type] = [module]
        for m in Reflect.get_metadata(module, ModuleMetadata.Imports, []):
            if isinstance(m, DynamicModule):
                modules.append(m.module)
            else:
                modules.append(m)
        return uniq_list(modules)

    def init(self, root_module: Type):
        self._root_module = root_module
        self._set_metadata()

    async def run(self, command_name: str, args: tuple):
        try:
            modules = self._get_modules(self._root_module)
            await self.instance_loader.create_instances(modules)
            commands: dict[str, BaseCommand] = {}
            for c in self.instance_loader.discover.get_all_provider():
                meta: Optional[dict[str]] = Reflect.get_metadata(c, CommanderMeta.Meta)
                if meta is not None and issubclass(c.__class__, BaseCommand):
                    commands.update({meta.get("name"): cast(BaseCommand, c)})

            command = commands.get(command_name, None)
            if command is not None:
                command.init(args)
                await command.run()
            else:
                echo.error(f"Command '{command_name}' not found ")
            await self.instance_loader.destroy()

        except Exception as e:
            _tb = traceback.format_exc()
            echo.error(e)
            echo.error(_tb)

    def _set_metadata(self):
        provider_metadata_maker = ProviderMetadataCreator(
            self._root_module, is_root=True
        )
        provider_metadata_maker.create()

        controller_metadata_maker = ControllerMetadataCreator(
            self._root_module, is_root=True
        )
        controller_metadata_maker.create()

        # optional
        module_metadata_maker = ModuleMetadataCreator(self._root_module)
        module_metadata_maker.create()



import inspect
import re
import traceback
from typing import Callable, Any, Optional

from nestipy.ioc import MiddlewareContainer, MiddlewareProxy

from nestipy.common.logger import logger
from nestipy.common.http_ import Request, Response
from nestipy.common.middleware import NestipyMiddleware
from nestipy.types_ import HTTPMethod


def uniq_middleware_list(data: list[MiddlewareProxy]) -> list:
    uniq_middleware: list = []
    uniq_data: list = []
    for d in data:
        if d.middleware not in uniq_middleware:
            uniq_data.append(d)
            uniq_middleware.append(d.middleware)
    return uniq_data


class MiddlewareExecutor:
    def __init__(self, req: Request, res: Response, next_fn: Callable):
        self.container = MiddlewareContainer.get_instance()
        # load all middleware inside a container
        self._middlewares: list[MiddlewareProxy] = self.container.all()
        self._req = req
        self._res = res
        self._next_fn = next_fn

    async def execute(self):
        middleware_to_apply = []
        for proxy in self._middlewares:
            if (
                self._is_match(proxy.route.url)
                and not self._is_exclude(proxy.path_excludes, proxy.route.url)
                and self._is_method_match(proxy.route.method)
            ):
                for p in proxy.middlewares:
                    p = MiddlewareProxy.form_dict(p, proxy.route, proxy.path_excludes)
                    self.container.add_singleton(p)
                    middleware_to_apply.append(p)
        # get all middleware that match request path
        if len(middleware_to_apply) == 0:
            # if no middleware call next_fn that call handler
            return await self._next_fn()
        middleware_to_apply = uniq_middleware_list(middleware_to_apply)
        return await self._recursively_call_middleware(0, middleware_to_apply)

    async def _create_middleware_callable(self, proxy: MiddlewareProxy):
        if inspect.isclass(proxy.middleware) and issubclass(
            proxy.middleware, NestipyMiddleware
        ):
            try:
                #  get instance of Middleware
                instance = await self.container.get(proxy)
                # get use method if it is a middleware class
                return getattr(instance, "use")
            except Exception as e:
                tb = traceback.format_exc()
                logger.error(e)
                logger.error(tb)
                return None
        elif inspect.isfunction(proxy.middleware):
            return proxy.middleware
        else:
            raise Exception(
                "Middleware must be a function or a class that extends NestipyMiddleware"
            )

    async def _recursively_call_middleware(
        self, index: int, middlewares: list[MiddlewareProxy]
    ) -> Any:
        current = middlewares[index]
        to_call = await self._create_middleware_callable(current)
        if index != len(middlewares) - 1:
            # create next_fn that cal next middleware
            async def next_fn():
                return await self._recursively_call_middleware(index + 1, middlewares)

            return await to_call(self._req, self._res, next_fn)
        else:
            return await to_call(self._req, self._res, self._next_fn)

    def _is_match(self, to_match: str, route: Optional[str] = None) -> bool:
        pattern = re.compile(f"^{to_match}")
        mitch = pattern.match(
            route or self._req.path,
        )
        return mitch is not None

    def _is_method_match(self, method: list[HTTPMethod]) -> bool:
        if "ALL" in method or "ANY" in method:
            return True
        else:
            return self._req.method.upper() in [m.upper() for m in method]

    def _is_exclude(self, excludes: list[str], to_match: str) -> bool:
        for ex in excludes:
            if self._is_match(to_match, ex):
                return True
        return False



import typing
from http import cookies as http_cookies
from typing import Callable, Optional, Any
from urllib.parse import parse_qsl

import ujson
from python_multipart.multipart import parse_options_header
from starlette._utils import AwaitableOrContextManagerWrapper, AwaitableOrContextManager
from starlette.requests import Request as StarletteRequest

from .form_parsers import MultiPartParser, MultiPartException, FormParser, FormData
from .multipart import parse_content_header


def cookie_parser(cookie_string: str) -> dict[str, str]:
    cookie_dict: dict[str, str] = {}
    for chunk in cookie_string.split(";"):
        if "=" in chunk:
            key, val = chunk.split("=", 1)
        else:
            key, val = "", chunk
        key, val = key.strip(), val.strip()
        if key or val:
            cookie_dict[key] = http_cookies._unquote(val)
    return cookie_dict


class ClientDisconnect(Exception):
    pass


class Request:
    def __init__(self, scope: dict, receive: Callable, send: Callable) -> None:
        self.scope = scope
        self.receive = receive
        self.send = receive
        self._query_params = None
        self._headers: Optional[dict] = None
        self._path = None
        self._method = None
        self._client = None
        self._host = None
        self._body = None
        self._json = None
        self._files = None
        self._user = None
        self._form = None
        self._session = None
        self._cookies = None
        self._stream_consumed = False
        self._is_disconnected = False
        if scope["type"] == "http":
            self.starlette_request = StarletteRequest(scope, receive, send)

    @property
    def query_params(self) -> dict:
        if self._query_params is None:
            self._query_params = {
                k.decode(): v.decode()
                for k, v in dict(parse_qsl(self.scope["query_string"])).items()
            }
        return self._query_params

    @property
    def path_params(self) -> dict[str, Any]:
        return self.scope.get("path_params", {})

    @property
    def path(self) -> str:
        if self._path is None:
            self._path = self.scope.get("raw_path").decode()
        return self._path

    @property
    def method(self) -> str:
        if self._method is None:
            self._method = self.scope.get("method")
        return self._method

    @property
    def user(self) -> Any:
        return self._user

    @user.setter
    def user(self, u: Any):
        self._user = u

    @property
    def host(self) -> str:
        if self._host is None:
            host: tuple[str, int] = self.scope.get("server")
            self._host = f"{self.scope.get('scheme')}://{host[0]}:{host[1]}"
        return self._host

    @property
    def headers(self) -> dict:
        if self._headers is None:
            self._headers = {}
            scope_headers: list[tuple[bytes, bytes]] = list(self.scope["headers"])
            for key, value in scope_headers:
                self._headers[key.decode()] = value.decode()
        return self._headers

    @property
    def client(self) -> list:
        if self._client is None:
            self._client = self.scope.get("client", ())
        return self._client

    @property
    def session(self) -> dict:
        if self._session is None:
            self._session = {}
        return self._session

    @session.setter
    def session(self, session: dict):
        self._session = session

    async def stream(self) -> typing.AsyncGenerator[bytes, None]:
        """
        Returns:
            stream: A stream generator value form request
        """
        if self._body is not None:
            yield self._body
            yield b""
            return
        if self._stream_consumed:
            raise RuntimeError("Stream consumed")
        while not self._stream_consumed:
            message = await self.receive()
            if message["type"] == "http.request":
                body = message.get("body", b"")
                if not message.get("more_body", False):
                    self._stream_consumed = True
                if body:
                    yield body
            elif message["type"] == "http.disconnect":
                self._is_disconnected = True
                raise ClientDisconnect()
        yield b""

    async def body(self) -> bytes:
        """
        Returns:
            body(bytes): Return bytes version of request body
        """
        if self._body is None:
            chunks: list[bytes] = []
            async for chunk in self.stream():
                chunks.append(chunk)
            self._body = b"".join(chunks)
        return self._body

    async def json(self) -> typing.Any:
        """

        Returns:
            json(dict): Return json value of request body
        """
        if self._json is None:
            body = await self.body()
            try:
                self._json = ujson.loads("{}" if body == "" else body)
            except ujson.JSONDecodeError:
                self._json = {}
        return self._json

    @property
    def content_type(self) -> tuple[str, dict[str, str]]:
        """

        Returns:
            content_type: Content-Type of request body
        """
        _content_type = self.headers.get("content-type", "")
        return parse_content_header(_content_type)

    def form(
        self, *, max_files: int | float = 1000, max_fields: int | float = 1000
    ) -> AwaitableOrContextManager[FormData]:
        return AwaitableOrContextManagerWrapper(
            self._get_form(max_files=max_files, max_fields=max_fields)
        )

    async def _get_form(
        self, *, max_files: int | float = 1000, max_fields: int | float = 1000
    ) -> FormData:
        if self._form is None:
            assert parse_options_header is not None, (
                "The `python-multipart` library must be installed to use form parsing."
            )
            content_type_header = self.headers.get("content-type")
            content_type: bytes
            content_type, _ = parse_options_header(content_type_header)
            if content_type == b"multipart/form-data":
                try:
                    multipart_parser = MultiPartParser(
                        self.headers,
                        self.stream(),
                        max_files=max_files,
                        max_fields=max_fields,
                    )
                    self._form = await multipart_parser.parse()
                except MultiPartException as exc:
                    raise exc
            elif content_type == b"application/x-www-form-urlencoded":
                form_parser = FormParser(self.headers, self.stream())
                self._form = await form_parser.parse()
            else:
                self._form = FormData()
        return self._form

    @property
    def cookies(self) -> dict[str, str]:
        if self._cookies is None:
            cookies: dict[str, str] = {}
            cookie_header = self.headers.get("cookie")
            if cookie_header:
                cookies = cookie_parser(cookie_header)
            self._cookies = cookies
        return self._cookies



from .handler import CallableHandler, NextFn, WebsocketHandler, MountHandler
from .http import HTTPMethod

__all__ = [
    "HTTPMethod",
    "CallableHandler",
    "WebsocketHandler",
    "MountHandler",
    "NextFn",
]



import uuid
from dataclasses import dataclass, field
from typing import (
    TypeVar,
    Generic,
    Optional,
    Any,
    Callable,
    Union,
    Type,
    Awaitable,
    Dict,
    List,
)

from nestipy.ioc.provider import ModuleProviderDict
from nestipy.metadata import ModuleMetadata, Reflect

T = TypeVar("T")


@dataclass
class DynamicModule:
    module: Any
    exports: list = field(default_factory=lambda: [])
    imports: list = field(default_factory=lambda: [])
    providers: list = field(default_factory=lambda: [])
    controllers: list = field(default_factory=lambda: [])
    is_global: bool = False


class ConfigurableModuleBuilder(Generic[T]):
    def __init__(self):
        self._method_name = "register"
        self._extras: Union[dict[str, Any], None] = None
        self._extras_process_callback: Union[
            Callable[[DynamicModule, dict[str, Any]], None], None
        ] = None

    def set_extras(
        self,
        extras: dict[str, Any],
        extras_callback: Callable[[DynamicModule, dict[str, Any]], None],
    ):
        self._extras = extras
        self._extras_process_callback = extras_callback
        return self

    def set_method(self, name: str):
        self._method_name = name
        return self

    def _extra_return(self, dynamic_module: DynamicModule) -> DynamicModule:
        if self._extras is not None and self._extras_process_callback is not None:
            self._extras_process_callback(dynamic_module, self._extras)
        return dynamic_module

    def _create_dynamic_module(
        self, obj: Any, imports: list, provider: list
    ) -> DynamicModule:
        dynamic_module = DynamicModule(
            obj,
            providers=provider
            + Reflect.get_metadata(obj, ModuleMetadata.Providers, []),
            exports=Reflect.get_metadata(obj, ModuleMetadata.Exports, []),
            imports=imports + Reflect.get_metadata(obj, ModuleMetadata.Imports, []),
            controllers=Reflect.get_metadata(obj, ModuleMetadata.Controllers, []),
            is_global=Reflect.get_metadata(obj, ModuleMetadata.Global, False),
        )
        return self._extra_return(dynamic_module)

    def build(self) -> tuple[Type, str]:
        module_option_token = f"{uuid.uuid4().hex}_TOKEN"

        def register(
            cls_: Any, options: Optional[T], extras: Optional[dict] = None
        ) -> DynamicModule:
            if extras is not None:
                self._extras = extras
            provider = ModuleProviderDict(token=module_option_token, value=options)
            return self._create_dynamic_module(cls_, [], [provider])

        def register_async(
            cls_: Any,
            value: Optional[T] = None,
            factory: Optional[Callable[..., Union[Awaitable, Any]]] = None,
            existing: Optional[Union[Type, str]] = None,
            use_class: Optional[Type] = None,
            inject: Optional[list] = None,
            imports: Optional[list] = None,
            extras: Optional[dict] = None,
        ) -> DynamicModule:
            if extras is not None:
                self._extras = extras
            provider = ModuleProviderDict(
                token=module_option_token,
                factory=factory,
                inject=inject or [],
                use_class=use_class,
                existing=existing,
                value=value,
                imports=imports,
            )
            return self._create_dynamic_module(cls_, imports or [], [provider])

        annotations = {
            self._method_name: Callable[
                [Any, Optional[T], Optional[Dict[str, Any]]], DynamicModule
            ],
            f"{self._method_name}_async": Callable[
                [
                    Any,
                    Optional[T],
                    Optional[Callable[..., Union[Awaitable[Any], Any]]],
                    Optional[Union[Type, str]],
                    Optional[Type],
                    Optional[List[Any]],
                    Optional[List[Any]],
                    Optional[Dict[str, Any]],
                ],
                DynamicModule,
            ],
        }

        cls = type(
            "ConfigurableModuleClass",
            (object,),
            {
                "__annotations__": annotations,
                self._method_name: classmethod(register),
                f"{self._method_name}_async": classmethod(register_async),
            },
        )
        return cls, module_option_token



from typing import Type, Union

from nestipy.metadata import SetMetadata
from .can_activate import CanActivate
from .meta import GuardKey


def UseGuards(*guards: Union[Type, CanActivate]):
    """
    Guard decorator
    Args:
        *guards(Union[Type, CanActivate]): List of guards.

    Returns:
        decorator(): UseGuards decorator
    """
    return SetMetadata(
        GuardKey.Meta,
        [g for g in guards if issubclass(g, CanActivate) or callable(g)],
        as_list=True,
    )



from .cors import cors
from .helmet import helmet
from .interface import NestipyMiddleware
from .meta import MiddlewareKey
from .session import (
    session,
    SessionOption,
    cookie_session,
    SessionStore,
    SessionCookieOption,
)

__all__ = [
    "cors",
    "NestipyMiddleware",
    "MiddlewareKey",
    "SessionOption",
    "session",
    "helmet",
    "cookie_session",
    "SessionStore",
    "SessionCookieOption",
]



class CtxDepKey:
    Service: str = "service"
    Request: str = "request"
    Response: str = "response"
    Session: str = "session"
    Sessions: str = "sessions"
    Cookie: str = "cookie"
    Cookies: str = "cookies"
    Header: str = "header"
    Headers: str = "headers"
    Query: str = "query"
    Queries: str = "queries"
    Body: str = "body"
    Params: str = "params"
    Param: str = "param"
    Args: str = "args"
    Arg: str = "arg"
    Context: str = "execution_context"
    Files: str = "files"
    WebSocketClient: str = "io_client"
    SocketData: str = "io_data"

    @classmethod
    def to_list(cls) -> list:
        return [
            cls.Service,
            cls.Request,
            cls.Param,
            cls.Response,
            cls.Session,
            cls.Queries,
            cls.Query,
            cls.Body,
            cls.Params,
            cls.Args,
            cls.Arg,
            cls.WebSocketClient,
            cls.Cookie,
            cls.Files,
            cls.Cookies,
            cls.Context,
            cls.Sessions,
            cls.SocketData,
            cls.Header,
            cls.Headers,
        ]



from typing import Callable

from nestipy.common import Request


class Websocket(Request):
    def __init__(
        self, namespace, sid, data, scope: dict, receive: Callable, send: Callable
    ):
        super().__init__(scope, receive, send)
        self.namespace = namespace
        self.sid = sid
        self.data = data



from dataclasses import asdict
from typing import Union, Callable, MutableMapping, Any, Awaitable

from starlette.requests import Request
from starlette.responses import HTMLResponse, PlainTextResponse, Response
from starlette.websockets import WebSocket
from strawberry.asgi import GraphQL
from strawberry.http.exceptions import HTTPException
from strawberry.http.typevars import Context
from strawberry.schema import BaseSchema
from strawberry.subscriptions import GRAPHQL_TRANSPORT_WS_PROTOCOL, GRAPHQL_WS_PROTOCOL
from strawberry.printer import print_schema
from ..graphql_asgi import GraphqlASGI
from ..graphql_module import GraphqlOption, ASGIOption


class StrawberryASGI(GraphQL, GraphqlASGI):
    def __init__(
        self,
        schema: BaseSchema,
        option: GraphqlOption,
    ):
        asgi_option = asdict(option.asgi_option or ASGIOption())
        asgi_option["subscription_protocols"] = asgi_option[
            "subscription_protocols"
        ] or (
            GRAPHQL_TRANSPORT_WS_PROTOCOL,
            GRAPHQL_WS_PROTOCOL,
        )

        super().__init__(schema=schema, **asgi_option)
        self.set_graphql_option(option)

    def print_schema(self) -> str:
        return print_schema(self.schema)

    async def get_context(
        self, request: Union[Request, WebSocket], response: Response
    ) -> Context:
        context = {"request": request, "response": response}
        return await self.modify_default_context(context)

    async def render_graphql_ide(self, request: Union[Request, WebSocket]) -> Response:
        if self.option.ide:
            return HTMLResponse(await self.get_graphql_ide())
        else:
            raise HTTPException(404, "Not Found")

    async def handle(self, scope: dict, receive, send):
        await super().handle(scope, receive, send)
        await self.__call__(scope, receive, send)

    async def handle_http(
        self,
        scope: MutableMapping[str, Any],
        receive: Callable[[], Awaitable[Any]],
        send: Callable[[...], Awaitable[None]],
    ) -> None:
        request = Request(scope=scope, receive=receive)
        try:
            response = await self.run(request)
        except HTTPException as e:
            response = PlainTextResponse(e.reason, status_code=e.status_code)
        # Apply cors to graphql
        if self.option and self.option.cors:
            response.headers.update(
                {
                    "access-control-allow-origin": "*",
                    "access-control-allow-headers": "Content-Type",
                    "access-control-allow-methods": "GET, POST, PUT, DELETE, OPTIONS",
                }
            )
        await response(scope, receive, send)



from typing import Optional, Any

import click


class CliStyle:
    @classmethod
    def info(
        cls,
        info: Optional[Any] = None,
    ):
        click.secho(info, fg="green")

    @classmethod
    def error(
        cls,
        info: Optional[Any] = None,
    ):
        click.secho(info, fg="red")

    @classmethod
    def warning(
        cls,
        info: Optional[Any] = None,
    ):
        click.secho(info, fg="orange")

    @classmethod
    def success(
        cls,
        info: Optional[Any] = None,
    ):
        click.secho(info, fg="green", bold=True)


echo = CliStyle()



import asyncio

from nestipy.microservice.client import MqttClientProxy, MicroserviceOption
from nestipy.microservice.client.option import MqttClientOption
from nestipy.microservice.server.base import MicroServiceServer


async def main():
    client = MqttClientProxy(
        MicroserviceOption(option=MqttClientOption(hostname="localhost", port=1883)),
    )
    server = MicroServiceServer(pub_sub=client)
    print("Running server ...")
    await server.listen()


if __name__ == "__main__":
    asyncio.run(main())



"""
This module provides methods to obtain OpenAPI Documentation from file or web sources.
"""

import json
from pathlib import Path

import yaml

from openapidocs.logs import logger

from .web import ensure_success, http_get


def read_from_json_file(file_path: Path):
    """
    Reads JSON from a given file by path.
    """
    with open(file_path, "rt", encoding="utf-8") as source_file:
        return json.loads(source_file.read())


def read_from_yaml_file(file_path: Path):
    """
    Reads YAML from a given file by path.
    """
    with open(file_path, "rt", encoding="utf-8") as source_file:
        return yaml.safe_load(source_file.read())


class SourceError(Exception):
    def __init__(self, message: str) -> None:
        super().__init__(message)


def read_from_url(url: str):
    """
    Tries to read OpenAPI Documentation from the given source URL.
    This method will try to fetch JSON or YAML from the given source, in case of
    ambiguity regarding the content, it will to parse anyway the response as JSON or
    YAML (using safe load when handling YAML).
    """
    response = http_get(url)

    ensure_success(response)

    data = response.text
    content_type = response.headers.get("content-type")

    if "json" in content_type or url.endswith(".json"):
        return json.loads(data)

    if "yaml" in content_type or url.endswith(".yaml") or url.endswith(".yml"):
        return yaml.safe_load(data)

    try:
        return json.loads(data)
    except json.JSONDecodeError:
        try:
            return yaml.safe_load(data)
        except yaml.YAMLError:
            raise SourceError(
                "Could not load a valid JSON or YAML file from the given URL."
            )


def read_from_source(source: str):
    """
    Tries to read a JSON or YAML file from a given source.
    The source can be a path to a file, or a URL.
    """
    source_path = Path(source)

    if source_path.exists():
        if not source_path.is_file():
            raise ValueError("The given path is not a file path.")

        logger.debug("Reading from file %s", source)

        file_path = source.lower()

        if file_path.endswith(".json"):
            return read_from_json_file(source_path)

        if file_path.endswith(".yaml") or file_path.endswith(".yml"):
            return read_from_yaml_file(source_path)

        raise ValueError("Unsupported source file.")
    else:
        source_lower = source.lower()

        if source_lower.startswith("http://") or source_lower.startswith("https://"):
            # fetch with a web request, read - ensure that it's JSON or YAML!
            return read_from_url(source)
        else:
            raise ValueError(
                "Invalid source: it must be either a path to a "
                ".json or .yaml file, or a valid URL."
            )



import inspect
from typing import Callable, Type, Any, get_type_hints, get_args, Optional

import strawberry
from strawberry import (
    type as strawberry_type,
    field as strawberry_field,
    mutation,
    subscription,
    Schema,
)
from strawberry.types.field import StrawberryField

from nestipy.ioc.dependency import TypeAnnotated
from nestipy.metadata.dependency import CtxDepKey
from ..graphql_adapter import GraphqlAdapter
from ..graphql_asgi import GraphqlASGI
from ..graphql_module import GraphqlOption
from ..strawberry.strawberry_asgi import StrawberryASGI


class StrawberryAdapter(GraphqlAdapter):
    def raise_exception(self, e: Exception):
        raise e

    _schema: Optional[Schema] = None

    def create_type_field_resolver(self, prop: dict, resolver: Callable) -> object:
        type_to_resolve: Type = prop["type"]
        prop_name = prop["name"]
        field = strawberry_field(resolver=resolver)
        field.name = prop_name
        fields: list[StrawberryField] = type_to_resolve.__strawberry_definition__.fields
        new_fields = [field if f.name == prop_name else f for f in fields]
        type_to_resolve.__strawberry_definition__.fields = new_fields
        return type_to_resolve

    def create_query_field_resolver(self, resolver: Callable) -> object:
        return strawberry_field(resolver)

    def create_mutation_field_resolver(self, resolver: Callable) -> object:
        return mutation(resolver)

    def create_subscription_field_resolver(self, resolver: Callable) -> object:
        return subscription(resolver)

    def modify_handler_signature(
        self, original_handler: Any, mutated_handler: Callable, default_return_type: Any
    ) -> Type:
        signature = inspect.signature(original_handler)
        return_annotation = (
            signature.return_annotation
            if (signature.return_annotation is not getattr(inspect, "_empty"))
            else default_return_type
        )
        # Set annotations for the new function
        if return_annotation is None:
            raise Exception(
                f"Method {original_handler.__name__} has no return annotation"
            )

        original_annotations = get_type_hints(original_handler)

        new_parameters = []
        for param_name, param in signature.parameters.items():
            args = get_args(param.annotation)
            if any(
                isinstance(arg, TypeAnnotated)
                and arg.metadata.key == CtxDepKey.Args
                and arg.metadata.token is None
                for arg in args
            ):
                new_parameters.append(param)

        new_parameters_keys = [k.name for k in new_parameters]
        # Add root and context
        if "root" not in new_parameters_keys and "info" not in new_parameters_keys:
            new_parameters.insert(
                0,
                inspect.Parameter(
                    "root", inspect.Parameter.POSITIONAL_OR_KEYWORD, annotation=Any
                ),
            )
            new_parameters.insert(
                1,
                inspect.Parameter(
                    "info",
                    inspect.Parameter.POSITIONAL_OR_KEYWORD,
                    annotation=strawberry.Info,
                ),
            )

        if (
            "self" in signature.parameters
            and list(signature.parameters.keys())[0] == "self"
        ):
            new_parameters.insert(0, signature.parameters["self"])
        elif (
            "cls" in signature.parameters
            and list(signature.parameters.keys())[0] == "cls"
        ):
            new_parameters.insert(0, signature.parameters["cls"])

        new_signature = inspect.Signature(
            parameters=new_parameters, return_annotation=return_annotation
        )
        mutated_handler.__annotations__ = original_annotations
        mutated_handler.__signature__ = new_signature
        return return_annotation

    def create_schema(self, **kwargs) -> Schema:
        query = self.create_query()
        mutation_ = self.create_mutation()
        subscription_ = self.create_subscription()
        if query is None:

            def health() -> bool:
                return True

            query = type("Query", (), {"health": strawberry_field(health)})
        return Schema(
            query=strawberry_type(query),
            mutation=strawberry_type(mutation_) if mutation_ is not None else None,
            subscription=strawberry_type(subscription_)
            if subscription_ is not None
            else None,
            **kwargs,
        )

    def create_graphql_asgi_app(
        self, schema: Any, option: GraphqlOption
    ) -> GraphqlASGI:
        app_asgi = StrawberryASGI(schema=schema, option=option)
        return app_asgi



from typing import TYPE_CHECKING

from nestipy.common.http_ import Response, Request
from .argument_host import ArgumentHost

if TYPE_CHECKING:
    from nestipy.common.http_ import Request, Response


class GraphqlArgumentHost(ArgumentHost):
    def get_request(self) -> Request:
        return self._req

    def get_response(self) -> Response:
        return self._res

    def get_args(self) -> dict:
        return self._graphql_args

    def get_context(self):
        return self._graphql_context



from .client import (
    ClientsModule,
    ClientsConfig,
    ClientModuleFactory,
    MicroserviceOption,
)
from .client import (
    RedisClientProxy,
    MqttClientProxy,
    ClientProxy,
    RabbitMQClientProxy,
    NatsClientProxy,
    TCPClientProxy,
    GrpcClientProxy,
)
from .client.option import (
    RedisClientOption,
    MqttClientOption,
    NatsClientOption,
    RabbitMQQueueOption,
    RabbitMQClientOption,
    TCPClientOption,
    GrpcClientOption,
)
from .client.option import Transport
from .context import RpcResponse, RpcRequest
from .decorator import MessagePattern, EventPattern
from .dependency import Payload, Ctx, Client, Context
from .exception import RpcException, RpcExceptionFilter, RPCErrorMessage, RPCErrorCode
from .serializer import JSONSerializer
from .server import MicroServiceServer

__all__ = [
    "ClientsModule",
    "ClientsConfig",
    "ClientProxy",
    "TCPClientProxy",
    "RedisClientProxy",
    "MqttClientProxy",
    "RabbitMQClientProxy",
    "NatsClientProxy",
    "GrpcClientProxy",
    "MicroserviceOption",
    "TCPClientOption",
    "RedisClientOption",
    "MqttClientOption",
    "RabbitMQQueueOption",
    "RabbitMQClientOption",
    "NatsClientOption",
    "GrpcClientOption",
    "Transport",
    "MessagePattern",
    "EventPattern",
    "RpcResponse",
    "RpcRequest",
    "RpcException",
    "RpcExceptionFilter",
    "MicroServiceServer",
    "Payload",
    "Ctx",
    "Context",
    "Client",
    "RPCErrorMessage",
    "RPCErrorCode",
    "ClientModuleFactory",
    "JSONSerializer",
]



import asyncio
import typing
from typing import AsyncIterator

import grpc

import nestipy.microservice.client.grpc.microservice_pb2 as pb2
import nestipy.microservice.client.grpc.microservice_pb2_grpc as pb2_grpc
from nestipy.microservice.client.base import ClientProxy, MicroserviceOption
from nestipy.microservice.client.option import GrpcClientOption
from nestipy.microservice.exception import RpcException, RPCErrorCode, RPCErrorMessage
from .server import GrpcServer


class GrpcClientProxy(ClientProxy):
    def __init__(self, option: MicroserviceOption):
        super().__init__(option)
        self.channel: typing.Optional[grpc.Channel] = None
        self.stub: typing.Optional[pb2_grpc.GrpcStub] = None
        self.subscription: typing.Optional[AsyncIterator[pb2.DataResponse]] = None
        self.response_queue = asyncio.Queue()
        self._config = typing.cast(
            GrpcClientOption, self.option.option or GrpcClientOption()
        )
        self.loop: asyncio.AbstractEventLoop = asyncio.get_running_loop()
        self.server: GrpcServer = GrpcServer()
        self.server_task: typing.Optional[asyncio.Task] = None

    async def slave(self) -> "ClientProxy":
        """Clone the client proxy."""
        return GrpcClientProxy(self.option)

    async def before_start(self):
        self.server_task = self.loop.create_task(
            coro=self.server.serve(port=self._config.port), name="grpc_server"
        )

    async def before_close(self):
        if self.server_task:
            self.server_task.cancel()

    async def connect(self):
        """Establish a connection to the gRPC server."""
        if not self.channel:
            self.channel = grpc.aio.insecure_channel(
                f"{self._config.host}:{self._config.port}"
            )
            self.stub = pb2_grpc.GrpcStub(self.channel)

    async def _publish(self, topic, data):
        """Publish data"""
        request = pb2.DataRequest(topic=topic, data=data)
        await self.stub.SendData(request)

    async def send_response(self, topic, data):
        """Send a response message."""
        await self._publish(topic, data)

    async def subscribe(self, topic: str, *args, **kwargs):
        sub_request = pb2.SubRequest(topic=topic)
        self.subscription = self.stub.Subscribe(sub_request)

    async def listen_response(self, from_topic: str, timeout: int = 30) -> str:
        """Listen for a response using an async queue."""
        try:
            async with asyncio.timeout(timeout):
                async for response in self.subscription:
                    if response.topic == from_topic:
                        return response.data
        except asyncio.TimeoutError:
            raise RpcException(
                status_code=RPCErrorCode.DEADLINE_EXCEEDED,
                message=RPCErrorMessage.DEADLINE_EXCEEDED,
            )

    async def listen(self) -> AsyncIterator[str]:
        """Continuously listen for incoming responses."""
        async for response in self.subscription:
            yield response.data

    async def unsubscribe(self, topic: str, *args):
        """Unsubscribe from topics (not applicable for gRPC)."""
        unsub_request = pb2.UnsubRequest(topic=topic)
        await self.stub.Unsubscribe(unsub_request)

    async def close(self):
        """Close the gRPC channel."""
        if self.channel:
            await self.channel.close()



from nestipy.common.http_ import Response, Request
from .argument_host import ArgumentHost


class WebsocketArgumentHost(ArgumentHost):
    def get_request(self) -> Request:
        return self._req

    def get_response(self) -> Response:
        return self._res

    def get_args(self) -> tuple[Request, Response]:
        return self.get_request(), self.get_response()

    def get_server(self):
        return self._socket_server

    def get_client(self):
        return self._socket_client

    def get_data(self):
        return self._socket_data



import asyncio
from typing import Any

import grpc
from rich.style import Style

import nestipy.microservice.client.grpc.microservice_pb2 as pb2
import nestipy.microservice.client.grpc.microservice_pb2_grpc as pb2_grpc
from nestipy.common.logger import console


class GrpcServer(pb2_grpc.GrpcServicer):
    def __init__(self, verbose: bool = True):
        self.subscriptions: dict[str, Any] = {}
        self.verbose: bool = verbose

    async def Subscribe(self, request, context):
        topic = request.topic
        if topic not in self.subscriptions:
            self.subscriptions[topic] = []
        self.subscriptions[topic].append(context)
        if self.verbose:
            console.print(
                f"INFO:     Client subscribed to {topic}",
                style=Style(color="green", bold=True),
            )
        try:
            while True:
                await asyncio.sleep(0.01)
        except grpc.aio.AioRpcError as e:
            if self.verbose:
                console.print(
                    f"ERROR:     Subscription error  {e}",
                    style=Style(color="red", bold=True),
                )
            if topic in self.subscriptions and context in self.subscriptions[topic]:
                self.subscriptions[topic].remove(context)
                if not self.subscriptions[topic]:
                    del self.subscriptions[topic]

    async def Unsubscribe(self, request, context):
        topic = request.topic
        if topic in self.subscriptions and context in self.subscriptions[topic]:
            self.subscriptions[topic].remove(context)
            if self.verbose:
                console.print(
                    f"INFO:     Client unsubscribe to {topic}",
                    style=Style(color="green", bold=True),
                )
            if not self.subscriptions[topic]:
                del self.subscriptions[topic]
        return pb2.Empty()

    async def SendData(self, request, context):
        topic = request.topic
        if self.verbose:
            console.print(
                f"INFO:     Client send data  to {topic}",
                style=Style(color="green", bold=True),
            )
        if topic in self.subscriptions:
            for client_context in self.subscriptions[topic]:
                try:
                    await client_context.write(
                        pb2.DataResponse(topic=topic, data=request.data)
                    )
                except grpc.aio.AioRpcError as e:
                    if self.verbose:
                        console.print(
                            f"ERROR:     Sending data error  {e}",
                            style=Style(color="red", bold=True),
                        )
                    self.subscriptions[topic].remove(client_context)
                    if not self.subscriptions[topic]:
                        del self.subscriptions[topic]
        return pb2.Empty()

    async def serve(self, port: int = 50051) -> None:
        """Start the server."""
        server = grpc.aio.server()
        pb2_grpc.add_GrpcServicer_to_server(self, server)
        server.add_insecure_port(f"[::]:{port}")
        console.print(
            f"INFO:     Grcp server running on port {port}.",
            style=Style(color="green", bold=True),
        )
        await server.start()
        await server.wait_for_termination()



from .http_adapter import HttpAdapter

try:
    from .blacksheep_adapter import BlackSheepAdapter
except ImportError:
    BlackSheepAdapter = None

from .fastapi_adapter import FastApiAdapter

# Dynamically build the __all__ list
__all__ = ["HttpAdapter"]
if BlackSheepAdapter:
    __all__.append("BlackSheepAdapter")
if FastApiAdapter:
    __all__.append("FastApiAdapter")



import asyncio
import enum
import logging
from dataclasses import dataclass, field
from ssl import SSLContext, TLSVersion
from typing import Union, Optional, Mapping, Literal, Iterable, List

import paho.mqtt.client as mqtt
from aio_pika.abc import SSLOptions, ExchangeType
from aiomqtt import ProtocolVersion, Will, Message, TLSParameters, ProxySettings
from aiomqtt.types import SocketOption, WebSocketHeaders
from aiormq.abc import TimeoutType
from nats.aio.client import (
    DEFAULT_CONNECT_TIMEOUT,
    DEFAULT_RECONNECT_TIME_WAIT,
    DEFAULT_MAX_RECONNECT_ATTEMPTS,
    DEFAULT_PING_INTERVAL,
    DEFAULT_MAX_OUTSTANDING_PINGS,
    DEFAULT_MAX_FLUSHER_QUEUE_SIZE,
    DEFAULT_DRAIN_TIMEOUT,
    Credentials,
    DEFAULT_INBOX_PREFIX,
    DEFAULT_PENDING_SIZE,
)
from paho.mqtt.properties import Properties
from pamqp.common import FieldTable, Arguments
from redis import ConnectionPool, CredentialProvider
from redis.asyncio.retry import Retry
from redis.utils import get_lib_version


class Transport(enum.Enum):
    TCP = "TCP"
    REDIS = "REDIS"
    MQTT = "MQTT"
    NATS = "NATS"
    RABBITMQ = "RABBITMQ"
    GRPC = "GRPC"
    CUSTOM = "CUSTOM"


@dataclass
class TCPClientOption:
    host: str = field(default="localhost")
    port: int = field(default=1333)
    start_server: bool = field(default=True)
    verbose: bool = field(default=True)


@dataclass
class GrpcClientOption:
    host: str = field(default="localhost")
    port: int = field(default=50051)
    verbose: bool = field(default=True)


@dataclass
class RedisClientOption:
    host: str = field(default="127.0.0.1")
    port: int = field(default=6379)
    db: Union[str, int] = 0
    password: Optional[str] = None
    socket_timeout: Optional[float] = None
    socket_connect_timeout: Optional[float] = None
    socket_keepalive: Optional[bool] = None
    socket_keepalive_options: Optional[Mapping[int, Union[int, bytes]]] = None
    connection_pool: Optional[ConnectionPool] = None
    unix_socket_path: Optional[str] = None
    encoding: str = "utf-8"
    encoding_errors: str = "strict"
    decode_responses: bool = True
    retry_on_timeout: bool = False
    retry_on_error: Optional[list] = None
    ssl: bool = False
    ssl_keyfile: Optional[str] = None
    ssl_certfile: Optional[str] = None
    ssl_cert_reqs: str = "required"
    ssl_ca_certs: Optional[str] = None
    ssl_ca_data: Optional[str] = None
    ssl_check_hostname: bool = False
    ssl_min_version: Optional[TLSVersion] = (None,)
    ssl_ciphers: Optional[str] = None
    max_connections: Optional[int] = None
    single_connection_client: bool = False
    health_check_interval: int = 0
    client_name: Optional[str] = None
    lib_name: Optional[str] = "redis-py"
    lib_version: Optional[str] = get_lib_version()
    username: Optional[str] = None
    retry: Optional[Retry] = None
    auto_close_connection_pool: Optional[bool] = None
    redis_connect_func = None
    credential_provider: Optional[CredentialProvider] = None
    protocol: Optional[int] = 2


@dataclass
class MqttClientOption:
    hostname: str = field(default="localhost")
    port: int = field(default=1883)
    username: Optional[str] = None
    password: Optional[str] = None
    logger: Optional[logging.Logger] = None
    identifier: Optional[str] = None
    queue_type: type[asyncio.Queue[Message]] | None = None
    protocol: Optional[ProtocolVersion] = None
    will: Optional[Will] = None
    clean_session: Optional[bool] = None
    transport: Literal["tcp", "websockets", "unix"] = "tcp"
    timeout: Optional[float] = None
    keepalive: int = 60
    bind_address: str = ""
    bind_port: int = 0
    clean_start: mqtt.CleanStartOption = mqtt.MQTT_CLEAN_START_FIRST_ONLY
    max_queued_incoming_messages: Optional[int] = None
    max_queued_outgoing_messages: Optional[int] = None
    max_inflight_messages: Optional[int] = None
    max_concurrent_outgoing_calls: Optional[int] = None
    properties: Optional[Properties] = None
    tls_context: Optional[mqtt.ssl.SSLContext] = None
    tls_params: Optional[TLSParameters] = None
    tls_insecure: Optional[bool] = None
    proxy: Optional[ProxySettings] = None
    socket_options: Optional[Iterable[SocketOption]] = None
    websocket_path: Optional[str] = None
    websocket_headers: Union[WebSocketHeaders, None] = None


@dataclass
class NatsClientOption:
    servers: Union[str, List[str]] = field(
        default_factory=lambda: ["nats://localhost:4222"]
    )
    name: Optional[str] = None
    pedantic: bool = False
    verbose: bool = False
    allow_reconnect: bool = True
    connect_timeout: int = DEFAULT_CONNECT_TIMEOUT
    reconnect_time_wait: int = DEFAULT_RECONNECT_TIME_WAIT
    max_reconnect_attempts: int = DEFAULT_MAX_RECONNECT_ATTEMPTS
    ping_interval: int = DEFAULT_PING_INTERVAL
    max_outstanding_pings: int = DEFAULT_MAX_OUTSTANDING_PINGS
    dont_randomize: bool = False
    flusher_queue_size: int = DEFAULT_MAX_FLUSHER_QUEUE_SIZE
    no_echo: bool = False
    tls: Optional[SSLContext] = None
    tls_hostname: Optional[str] = None
    tls_handshake_first: bool = False
    user: Optional[str] = None
    password: Optional[str] = None
    token: Optional[str] = None
    drain_timeout: int = DEFAULT_DRAIN_TIMEOUT
    user_credentials: Optional[Credentials] = None
    nkeys_seed: Optional[str] = None
    nkeys_seed_str: Optional[str] = None
    inbox_prefix: Union[str, bytes] = DEFAULT_INBOX_PREFIX
    pending_size: int = DEFAULT_PENDING_SIZE
    flush_timeout: Optional[float] = None


@dataclass
class RabbitMQQueueOption:
    name: str = None
    type: Union[ExchangeType, str] = ExchangeType.FANOUT
    durable: bool = False
    auto_delete: bool = False
    internal: bool = False
    passive: bool = False
    arguments: Arguments = None
    timeout: TimeoutType = None
    robust: bool = True


@dataclass
class RabbitMQClientOption:
    host: str = field(default="localhost")
    port: int = field(default=1883)
    login: str = field(default="guest")
    password: str = field(default="guest")
    virtualhost: str = field(default="/")
    ssl: bool = field(default=False)
    loop: Optional[asyncio.AbstractEventLoop] = None
    ssl_options: Optional[SSLOptions] = None
    ssl_context: Optional[SSLContext] = None
    timeout: TimeoutType = None
    client_properties: Optional[FieldTable] = None
    queue_option: RabbitMQQueueOption = field(
        default_factory=lambda: RabbitMQQueueOption()
    )



from dataclasses import dataclass, field
from typing import Union, Type, Callable, Literal, Optional

from nestipy.metadata import Reflect, RouteKey, ClassMetadata
from .container import NestipyContainer

HTTPMethod = Literal[
    "GET", "POST", "PUT", "DELETE", "PATCH", "HEAD", "OPTIONS", "ALL", "ANY"
]


@dataclass
class MiddlewareRouteConfig:
    method: list[HTTPMethod] = field(default_factory=lambda: ["ALL"])
    url: str = field(default="/")


class MiddlewareProxy:
    def __init__(self, *middleware: Union[Type, Callable]):
        self.middlewares = list(middleware)
        self.middleware = None
        self.path_excludes = []
        self.route: MiddlewareRouteConfig = MiddlewareRouteConfig()

    @classmethod
    def form_dict(
        cls,
        middleware: Union[Type, Callable],
        route: MiddlewareRouteConfig,
        path_excludes=None,
    ) -> "MiddlewareProxy":
        if path_excludes is None:
            path_excludes = []
        m = MiddlewareProxy()
        m.middleware = middleware
        m.route = route
        m.path_excludes = path_excludes
        return m

    def for_route(
        self, route: Union[Type, str], method: list[HTTPMethod] = None
    ) -> "MiddlewareProxy":
        self.route.method = method or ["ALL"]
        if isinstance(route, str):
            self.route.url = route
        else:
            self.route.url = (
                f"/{Reflect.get_metadata(route, RouteKey.path, '').strip('/')}"
            )
        return self

    def excludes(self, pattern=None) -> "MiddlewareProxy":
        if pattern is None:
            pattern = []
        if isinstance(pattern, str):
            self.path_excludes = [pattern]
        else:
            self.path_excludes = pattern
        return self


class MiddlewareContainer:
    _instance: Optional["MiddlewareContainer"] = None
    _middlewares: list = []
    _middleware_instances: dict = {}

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super(MiddlewareContainer, cls).__new__(
                cls, *args, **kwargs
            )
        return cls._instance

    @classmethod
    def get_instance(cls, *args, **kwargs):
        return MiddlewareContainer(*args, **kwargs)

    def add_singleton(
        self, proxy: MiddlewareProxy, module: Union[Type, object, None] = None
    ):
        if module is not None:
            for m in list(proxy.middlewares):
                Reflect.set_metadata(
                    m,
                    ClassMetadata.Metadata,
                    ClassMetadata(module=module, global_providers=[]),
                )
        self._middlewares.append(proxy)

    def all(self) -> list[MiddlewareProxy]:
        return self._middlewares

    async def get(self, proxy: MiddlewareProxy):
        if proxy in self._middlewares:
            if proxy in self._middleware_instances:
                return self._middleware_instances[proxy.middleware]
            else:
                instance = await NestipyContainer().get_instance().get(proxy.middleware)
                self._middleware_instances[proxy.middleware] = instance
                return instance
        else:
            raise ValueError(f"Middleware for route {proxy.route} not found")



from inspect import getmembers, isfunction
from typing import Union, Type

from nestipy.graphql.meta import NestipyGraphqlKey
from nestipy.metadata import ModuleMetadata, Reflect


class GraphqlExplorer:
    @classmethod
    def explore(
        cls, module_ref: Union[Type, object]
    ) -> tuple[list[dict], list[dict], list[dict], list[dict]]:
        query: list[dict] = []
        mutation: list[dict] = []
        subscription: list[dict] = []
        field_resolver: list[dict] = []
        providers = Reflect.get_metadata(module_ref, ModuleMetadata.Providers, [])
        for provider in providers:
            if not Reflect.get_metadata(provider, NestipyGraphqlKey.resolver, False):
                continue
            cls._explore_resolver(
                provider, query, mutation, subscription, field_resolver
            )

        return query, mutation, subscription, field_resolver

    @classmethod
    def _explore_resolver(
        cls,
        resolver: Type,
        query: list,
        mutation: list,
        subscription: list,
        field_resolver: list[dict],
    ):
        class_to_resolve = Reflect.get_metadata(
            resolver, NestipyGraphqlKey.return_type, None
        )
        members = getmembers(resolver, isfunction)
        for method_name, _ in members:
            if method_name.startswith("__"):
                continue
            method = getattr(resolver, method_name)
            graphql_handler = Reflect.get_metadata(
                method, NestipyGraphqlKey.handler, None
            )
            if graphql_handler is not None:
                return_type = Reflect.get_metadata(
                    method, NestipyGraphqlKey.return_type, None
                )
                data: dict = {
                    "return_type": return_type,
                    "class": resolver,
                    "handler_name": method_name,
                }
                match graphql_handler:
                    case NestipyGraphqlKey.query:
                        query.append(data)
                    case NestipyGraphqlKey.mutation:
                        mutation.append(data)
                    case NestipyGraphqlKey.subscription:
                        subscription.append(data)
                    case NestipyGraphqlKey.field_resolver:
                        if class_to_resolve is not None:
                            kwargs: dict = Reflect.get_metadata(
                                method, NestipyGraphqlKey.kwargs, {}
                            )
                            name: str = method_name
                            if "name" in kwargs and kwargs.get("name"):
                                name = kwargs.get("name")
                            data["name"] = name
                            data["type"] = class_to_resolve
                            field_resolver.append(data)



from .adapter import SocketIoAdapter, IoAdapter
from .socket_request import Websocket
from .decorator import Gateway, SuccessEvent, SubscribeMessage, ErrorEvent

__all__ = [
    "SocketIoAdapter",
    "IoAdapter",
    "Gateway",
    "SuccessEvent",
    "SubscribeMessage",
    "ErrorEvent",
    "Websocket",
]



from dataclasses import dataclass
from typing import Any, cast

from nestipy.dynamic_module import ConfigurableModuleBuilder, DynamicModule


@dataclass
class EventOption:
    pass


def extract_callback(dynamic_module: DynamicModule, extras: dict[str, Any]):
    if extras.get("is_global"):
        dynamic_module.is_global = cast(bool, extras.get("is_global", False))


ConfigurableClassBuilder, EVENT_CONFIG = (
    ConfigurableModuleBuilder[EventOption]().set_extras({}, extract_callback).build()
)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: microservice.proto
# Protobuf Python Version: 5.28.1
"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC, 5, 28, 1, "", "microservice.proto"
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x12microservice.proto\x12\x02ms"\x1b\n\nSubRequest\x12\r\n\x05topic\x18\x01 \x01(\t"\x1d\n\x0cUnsubRequest\x12\r\n\x05topic\x18\x01 \x01(\t"*\n\x0b\x44\x61taRequest\x12\r\n\x05topic\x18\x01 \x01(\t\x12\x0c\n\x04\x64\x61ta\x18\x02 \x01(\t"+\n\x0c\x44\x61taResponse\x12\x0c\n\x04\x64\x61ta\x18\x01 \x01(\t\x12\r\n\x05topic\x18\x02 \x01(\t"\x07\n\x05\x45mpty2\x8b\x01\n\x04Grpc\x12/\n\tSubscribe\x12\x0e.ms.SubRequest\x1a\x10.ms.DataResponse0\x01\x12*\n\x0bUnsubscribe\x12\x10.ms.UnsubRequest\x1a\t.ms.Empty\x12&\n\x08SendData\x12\x0f.ms.DataRequest\x1a\t.ms.Emptyb\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "microservice_pb2", _globals)
if not _descriptor._USE_C_DESCRIPTORS:
    DESCRIPTOR._loaded_options = None
    _globals["_SUBREQUEST"]._serialized_start = 26
    _globals["_SUBREQUEST"]._serialized_end = 53
    _globals["_UNSUBREQUEST"]._serialized_start = 55
    _globals["_UNSUBREQUEST"]._serialized_end = 84
    _globals["_DATAREQUEST"]._serialized_start = 86
    _globals["_DATAREQUEST"]._serialized_end = 128
    _globals["_DATARESPONSE"]._serialized_start = 130
    _globals["_DATARESPONSE"]._serialized_end = 173
    _globals["_EMPTY"]._serialized_start = 175
    _globals["_EMPTY"]._serialized_end = 182
    _globals["_GRPC"]._serialized_start = 185
    _globals["_GRPC"]._serialized_end = 324
# @@protoc_insertion_point(module_scope)



from __future__ import annotations

import asyncio
import functools
import sys
import typing
import uuid
from asyncio import Task
from enum import Enum

if sys.version_info >= (3, 10):
    from typing import ParamSpec
else:
    from typing_extensions import ParamSpec

from starlette.concurrency import run_in_threadpool

P = ParamSpec("P")


class TaskStatus(str, Enum):
    PENDING = "pending"
    STARTED = "started"
    SUCCESS = "success"
    FAILED = "failed"
    CANCELLED = "cancelled"


def is_async_callable(obj: typing.Any) -> typing.Any:
    while isinstance(obj, functools.partial):
        obj = obj.func
    return asyncio.iscoroutinefunction(obj) or (
        callable(obj) and asyncio.iscoroutinefunction(getattr(obj, "__call__", None))
    )


class BackgroundTask:
    def __init__(
        self,
        func: typing.Callable[P, typing.Any],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> None:
        self.task_id = str(uuid.uuid4())
        self.func = func
        self.args = args
        self.kwargs = kwargs
        self.is_async = is_async_callable(func)
        self.status: TaskStatus = TaskStatus.PENDING
        self.error: typing.Optional[str] = None

    async def __call__(self) -> None:
        self.status = TaskStatus.STARTED
        try:
            if self.is_async:
                await self.func(*self.args, **self.kwargs)
            else:
                await run_in_threadpool(self.func, *self.args, **self.kwargs)
            self.status = TaskStatus.SUCCESS
        except asyncio.CancelledError:
            self.status = TaskStatus.CANCELLED
        except Exception as e:
            self.status = TaskStatus.FAILED
            self.error = str(e)
            print(f"Task {self.task_id} failed: {e}")


class BackgroundTasks:
    def __init__(self) -> None:
        self.queue = asyncio.Queue()
        self.is_running = False
        self.worker_task: typing.Optional[Task] = None
        self.tasks: dict[str, BackgroundTask] = {}

    def add_task(
        self,
        func: typing.Callable[P, typing.Any],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> str:
        task = BackgroundTask(func, *args, **kwargs)
        return self.add(task)

    def add(self, task: BackgroundTask) -> str:
        task_id = task.task_id
        self.tasks[task_id] = task
        self.queue.put_nowait(task)
        return task_id

    async def _worker(self) -> None:
        while self.is_running:
            task = await self.queue.get()
            try:
                await task()
            finally:
                self.queue.task_done()
            await asyncio.sleep(0.1)

    def run(self) -> None:
        if not self.is_running:
            self.is_running = True
            self.worker_task = asyncio.get_running_loop().create_task(self._worker())

    async def shutdown(self) -> None:
        self.is_running = False
        if self.worker_task:
            self.worker_task.cancel()
        await self.queue.join()

    def get_status(self, task_id: str) -> typing.Optional[dict[str, typing.Any]]:
        task = self.tasks.get(task_id)
        if task:
            return {
                "id": task.task_id,
                "status": task.status,
                "error": task.error,
            }
        return {
            "id": task_id,
            "status": TaskStatus.CANCELLED,
            "error": "Task not found",
        }

    def all_tasks(self) -> list[dict[str, typing.Any]]:
        return [
            {
                "id": task.task_id,
                "status": task.status,
                "error": task.error,
            }
            for task in self.tasks.values()
        ]



import asyncio

from nestipy.microservice.client import NatsClientProxy, MicroserviceOption
from nestipy.microservice.client.option import NatsClientOption, Transport


async def main():
    client = NatsClientProxy(
        MicroserviceOption(transport=Transport.NATS, option=NatsClientOption()),
    )
    await client.send("topic", "request")


if __name__ == "__main__":
    asyncio.run(main())



import asyncio

from nestipy.microservice import Transport
from nestipy.microservice.client import TCPClientProxy, MicroserviceOption


async def main():
    client = TCPClientProxy(
        option=MicroserviceOption(transport=Transport.TCP),
    )
    response = await client.send("test", "request")
    print("Response from server:", response.data)


if __name__ == "__main__":
    asyncio.run(main())



import enum


class NestipyContextProperty(enum.Enum):
    request: str = "_request"
    response: str = "_response"
    query_params: str = "_query_params"
    params: str = "_params"
    session: str = "_session"
    cookies: str = "_cookies"
    body: str = "_body"
    headers: str = "_headers"
    args: str = "_args"
    files: str = "_files"
    execution_context: str = "_execution_context"
    io_server: str = "_websocket_sever"
    io_client: str = "_socket_sid"
    io_data: str = "_socket_data"



import logging
import typing
from importlib import import_module
from typing import Type

from .nestipy_application import NestipyApplication, NestipyConfig

from nestipy.microservice.client.base import MicroserviceOption
from .nestipy_microservice import NestipyMicroservice, NestipyConnectMicroservice


class _NestipyFactoryMeta(type):
    def __getitem__(self, item) -> "NestipyFactory":
        setattr(self, "__generic_type__", item.__name__)
        return typing.cast(NestipyFactory, self)


class NestipyFactory(metaclass=_NestipyFactoryMeta):
    @classmethod
    def create(
        cls, module: Type, config: typing.Optional[NestipyConfig] = None
    ) -> NestipyApplication:
        cls._setup_log()
        if getattr(cls, "__generic_type__", None) == "NestipyBlackSheepApplication":
            if not config:
                config = NestipyConfig(adapter=cls._load_adapter())
        application = NestipyApplication(config=config)
        application.init(module)
        return application

    @classmethod
    def create_microservice(
        cls, module: Type, option: list[MicroserviceOption]
    ) -> NestipyMicroservice:
        return NestipyMicroservice(module, option)

    @classmethod
    def connect_microservice(
        cls,
        module: Type,
        option: list[MicroserviceOption],
        config: typing.Optional[NestipyConfig] = None,
    ) -> NestipyConnectMicroservice:
        return NestipyConnectMicroservice(module, config, option)

    @classmethod
    def _setup_log(cls):
        logging.addLevelName(logging.INFO, "[NESTIPY] INFO")
        logging.addLevelName(logging.ERROR, "[NESTIPY] ERROR")
        logging.addLevelName(logging.WARNING, "[NESTIPY] WARNING")
        logging.addLevelName(logging.WARN, "[NESTIPY] WARN")
        logging.addLevelName(logging.CRITICAL, "[NESTIPY] CRITICAL")

    @classmethod
    def _load_adapter(cls):
        """
        Dynamically load the appropriate adapter (BlackSheep or FastAPI).
        """
        try:
            # Try to load the BlackSheep adapter
            adapter_module = import_module("nestipy.adapters.blacksheep_adapter")
            return adapter_module.BlackSheepAdapter()
        except ImportError:
            try:
                # Fallback to FastAPI adapter
                adapter_module = import_module("nestipy.adapters.fastapi_adapter")
                return adapter_module.FastAPIAdapter()
            except ImportError:
                raise RuntimeError(
                    "No suitable adapter found. Install extras with "
                    "`pip install nestipy[blacksheep]` or `pip install nestipy[fastapi]`."
                )



"""
This module provides a Jinja2 environment.
"""

import os
from enum import Enum

from jinja2 import Environment, PackageLoader, Template, select_autoescape

from . import get_http_status_phrase, highlight_params, read_dict, sort_dict
from .common import DocumentsWriter, is_reference
from .md import normalize_link, write_table


def configure_filters(env: Environment):
    env.filters.update(
        {"route": highlight_params, "table": write_table, "link": normalize_link}
    )


def configure_functions(env: Environment):
    helpers = {
        "read_dict": read_dict,
        "sort_dict": sort_dict,
        "is_reference": is_reference,
        "scalar_types": {"string", "integer", "boolean", "number"},
        "get_http_status_phrase": get_http_status_phrase,
        "write_md_table": write_table,
    }

    env.globals.update(helpers)


class OutputStyle(Enum):
    """
    Output style.
    """

    MKDOCS = 1
    """Markdown for MkDocs and PyMdown extensions"""

    MARKDOWN = 2
    """Basic Markdown"""

    PLANTUML_SCHEMAS = 100
    """PlantUML class diagram for components schemas."""

    PLANTUML_API = 101
    """PlantUML diagram of the API with request and response bodies."""


class PackageLoadingError(ValueError):
    def __init__(
        self, style: OutputStyle, templates_folder: str
    ) -> None:  # pragma: no cover
        super().__init__(
            f"Failed to read the templates for the output style {style.name}. "
            f"Tried to read templates from the folder: {__name__}.{templates_folder}. "
            "This is most probably an issue in `essentials-openapi`."
        )
        self.desired_style = style
        self.attempted_folder = templates_folder


def get_environment(
    package_name: str, views_style: OutputStyle = OutputStyle.MKDOCS
) -> Environment:
    templates_folder = f"views_{views_style.name}".lower()

    try:
        loader = PackageLoader(package_name, templates_folder)
    except ValueError as package_loading_error:  # pragma: no cover
        raise PackageLoadingError(
            views_style, templates_folder
        ) from package_loading_error
    else:
        env = Environment(
            loader=loader,
            autoescape=select_autoescape(["html", "xml"])
            if os.environ.get("SELECT_AUTOESCAPE") in {"YES", "Y", "1"}
            else False,
            auto_reload=True,
            enable_async=False,
        )
        configure_filters(env)
        configure_functions(env)

        return env


class Jinja2DocumentsWriter(DocumentsWriter):
    """
    This class uses Jinja2 templating engine to generate other kinds of text output from
    source OpenAPI Documentation data.
    """

    def __init__(
        self,
        package_name: str,
        views_style: OutputStyle = OutputStyle.MKDOCS,
    ) -> None:
        self._env = get_environment(package_name, views_style)

    @property
    def env(self) -> Environment:
        return self._env

    def get_template(self) -> Template:
        return self.env.get_template("layout.html")

    def write(self, data, **kwargs) -> str:
        template = self.get_template()
        return template.render(data, **kwargs)



from dataclasses import asdict
from typing import Literal, Annotated

from nestipy.common.decorator import Module, Injectable
from nestipy.dynamic_module import ConfigurableModuleBuilder
from nestipy.ioc import ModuleProviderDict, Inject
from nestipy.metadata import Reflect

Config = dict[Literal["folder"], str]
ConfigurableModuleClass, CONFIG_MODULE_OPTION_TOKEN = (
    ConfigurableModuleBuilder[Config]().set_method("for_root").build()
)


@Injectable()
class ConfigService:
    token: Annotated[str, Inject(CONFIG_MODULE_OPTION_TOKEN)]


@Module(
    providers=[ConfigService, ModuleProviderDict(value="hello", token="MyProvider")],
    exports=[ConfigService],
    is_global=True,
)
class ConfigModule(ConfigurableModuleClass): ...


if __name__ == "__main__":
    d = ConfigModule.for_root({"folder": "./config"})
    meta = Reflect.get(d)
    print(meta, {k: v for k, v in asdict(d).items() if k != "module"})

__all__ = ["ConfigService", "ConfigModule"]



import os.path
import traceback
from typing import Union, Type, Callable

from nestipy.core.guards import GuardProcessor
from nestipy.graphql.graphql_adapter import GraphqlAdapter
from nestipy.graphql.meta import NestipyGraphqlKey
from nestipy.ioc import NestipyContainer
from nestipy.ioc import RequestContextContainer
from nestipy.metadata import Reflect
from .graphql_explorer import GraphqlExplorer
from .graphql_module import GraphqlModule, GraphqlOption
from ..common import Request, Response, HttpException, HttpStatus, HttpStatusMessages
from ..common.logger import logger
from ..core.adapter.http_adapter import HttpAdapter
from ..core.context.execution_context import ExecutionContext
from ..types_ import NextFn


class GraphqlProxy:
    def __init__(self, adapter: HttpAdapter, graphql_server: GraphqlAdapter):
        self._graphql_server = graphql_server
        self._adapter = adapter
        self.container = NestipyContainer.get_instance()

    async def apply_resolvers(
        self, graphql_module: GraphqlModule, modules: list[Union[Type, object]]
    ):
        for module_ref in modules:
            query, mutation, subscription, field_resolver = GraphqlExplorer.explore(
                module_ref
            )
            for q in query:
                name, resolver = self._create_graphql_query_handler(module_ref, q)
                self._graphql_server.add_query_property(name, resolver)
            for m in mutation:
                name, resolver = self._create_graphql_query_handler(module_ref, m)
                self._graphql_server.add_mutation_property(name, resolver)
            for s in subscription:
                name, resolver = self._create_graphql_query_handler(module_ref, s)
                self._graphql_server.add_subscription_property(name, resolver)

            for fr in field_resolver:
                name, resolver = self._create_graphql_query_handler(
                    module_ref, fr, is_field=True
                )
                self._graphql_server.create_type_field_resolver(fr, resolver)

        self._create_graphql_request_handler(graphql_module.config or GraphqlOption())

    def _create_graphql_query_handler(
        self, module_ref: Union[Type, object], meta: dict, is_field: bool = False
    ) -> tuple[str, Callable]:
        resolver: Union[object, Type] = meta["class"]
        method_name: str = meta["handler_name"]
        method = getattr(resolver, method_name)
        default_return_type = Reflect.get_metadata(
            method, NestipyGraphqlKey.return_type, None
        )

        async def graphql_handler(*_args, **kwargs):
            context_container = RequestContextContainer.get_instance()
            default_context = context_container.execution_context
            execution_context = ExecutionContext(
                self._adapter,
                module_ref,
                resolver,
                getattr(resolver, method_name),
                default_context.get_request(),
                default_context.get_response(),
                {**default_context.switch_to_graphql().get_args(), **kwargs}
                if is_field
                else kwargs,
                None,
            )
            context_container.set_execution_context(execution_context)
            try:
                # TODO: Refactor with routerProxy
                # create execution context
                #  apply guards
                guard_processor: GuardProcessor = await self.container.get(
                    GuardProcessor
                )
                can_activate = await guard_processor.process(execution_context)
                if not can_activate[0]:
                    # TODO: is this need to specify return Exception
                    raise HttpException(
                        HttpStatus.UNAUTHORIZED, HttpStatusMessages.UNAUTHORIZED
                    )
                # perform query request
                result = await self.container.get(resolver, method_name)
            except Exception as e:
                tb = traceback.format_exc()
                logger.error(e)
                logger.error(tb)
                result = self._graphql_server.raise_exception(e)
            if is_field:
                context_container.set_execution_context(default_context)
            else:
                context_container.destroy()
            return result

        # mutate handle inside adapter
        self._graphql_server.modify_handler_signature(
            method, graphql_handler, default_return_type
        )
        return (
            method_name,
            graphql_handler,
        )

    def _create_graphql_request_handler(self, option: GraphqlOption):
        graphql_path = f"/{option.url.strip('/')}"
        gql_asgi = self._graphql_server.create_graphql_asgi_app(
            option=option,
            schema=self._graphql_server.create_schema(**(option.schema_option or {})),
        )
        if option.auto_schema_file:
            schema_sdl = gql_asgi.print_schema()
            with open(os.path.join(os.getcwd(), option.auto_schema_file), "w+") as file:
                file.write(schema_sdl)
        # create graphql handler but using handle of graphql_adapter
        self._adapter.mount(graphql_path, gql_asgi.handle)

        async def graphql_redirect(_req: Request, res: Response, _next_fn: NextFn):
            return await res.redirect(f"{graphql_path}/")

        self._adapter.get(graphql_path, graphql_redirect, {})

    @classmethod
    def should_render_graphql_ide(cls, req: Request) -> bool:
        return (
            req.method == "GET"
            and req.query_params.get("query") is None
            and any(
                supported_header in req.headers.get("accept", "")
                for supported_header in ("text/html", "*/*")
            )
        )



"""
This module provides common functions to generate other kinds of representations from
source OpenAPI Documentation files.
"""

import re
from http import HTTPStatus

import markupsafe


def get_http_status_phrase(status_code) -> str:
    try:
        http_status = HTTPStatus(int(status_code))
        return http_status.phrase
    except ValueError:
        return ""


def read_dict(obj, *args, default=None):
    """
    Reads properties in a source dictionary, returning None if any

    Example:
    read_dict({"a": {"b": {"c": True}}}, "a", "b", "c") --> True
    """
    assert isinstance(obj, dict)

    value = obj
    for key in args:
        if not isinstance(value, dict):
            raise ValueError(f"Invalid sub-path: {repr(args)}")

        value = value.get(key)

        if value is None:
            return default

    if value is None or value is obj:
        return default

    return value


def sort_dict(obj):
    """
    Yields (key, value) of a dictionary with keys in alphabetical order.
    """
    for key in sorted(obj, key=str.lower):
        yield key, obj[key]


def highlight_params(path: str) -> str:
    def replacer(match):
        value = match.group()
        return f'<span class="route-param">{markupsafe.escape(value)}</span>'

    return re.sub(r"\{[^\}]+\}", replacer, path)



from typing import Type, List, Optional

from pydantic import BaseModel, Field

from nestipy.dynamic_module import ConfigurableModuleBuilder


class RouteItem(BaseModel):
    path: str
    module: Type
    children: Optional[List["RouteItem"]] = Field(
        default_factory=lambda: [], alias="children"
    )


ConfigurableClassBuilder, ROUTER_CONFIG = (
    ConfigurableModuleBuilder[list[RouteItem]]().set_method("_register").build()
)



from dataclasses import dataclass

from nestipy.common.decorator.class_ import Module
from nestipy.dynamic_module import ConfigurableModuleBuilder, DynamicModule
from nestipy.ioc.provider import ModuleProviderDict
from .base import MicroserviceOption
from .factory import ClientModuleFactory


@dataclass
class ClientsConfig:
    name: str
    option: MicroserviceOption


ConfigurableModuleClass, CLIENT_OPTION = ConfigurableModuleBuilder[
    list[ClientsConfig]
]().build()


@Module()
class ClientsModule(ConfigurableModuleClass):
    @classmethod
    def register(cls, option: list[ClientsConfig]):
        dynamic_module: DynamicModule = getattr(ConfigurableModuleClass, "register")(
            option
        )
        providers = []
        for opt in option:
            providers.append(
                ModuleProviderDict(
                    token=opt.name, value=ClientModuleFactory.create(opt.option)
                )
            )
        dynamic_module.providers = dynamic_module.providers + providers
        dynamic_module.is_global = True
        return dynamic_module



from typing import Callable, Literal, Union

from nestipy.metadata import Reflect, RouteKey

HTTPMethod = Literal[
    "GET", "POST", "PUT", "DELETE", "PATCH", "HEAD", "OPTIONS", "ALL", "ANY"
]


class Route:
    def __init__(
        self, path: str = "", method: Union[list[HTTPMethod], None] = None, **kwargs
    ):
        if method is None:
            method = ["GET"]
        self.path = path
        self.kwargs = kwargs
        self.method = method

    def __call__(self, handler: Callable):
        # put path and kwargs in controller handler
        Reflect.set_metadata(handler, RouteKey.path, self.path)
        Reflect.set_metadata(handler, RouteKey.kwargs, self.kwargs)
        Reflect.set_metadata(handler, RouteKey.method, self.method)
        return handler


def Get(path: str = "", **kwargs):
    return Route(path=path, method=["GET"], **kwargs)


def Post(path: str = "", **kwargs):
    return Route(path=path, method=["POST"], **kwargs)


def Put(path: str = "", **kwargs):
    return Route(path=path, method=["PUT"], **kwargs)


def Patch(path: str = "", **kwargs):
    return Route(path=path, method=["PATCH"], **kwargs)


def Delete(path: str = "", **kwargs):
    return Route(path=path, method=["DELETE"], **kwargs)



from abc import ABC, abstractmethod
from typing import Any, Callable, Type, Union

from .graphql_asgi import GraphqlASGI
from .graphql_module import GraphqlOption


class GraphqlAdapter(ABC):
    _query_properties: list = []
    _mutation_properties: list = []
    _subscription_properties: list = []

    @abstractmethod
    def create_query_field_resolver(self, resolver: Callable) -> object:
        pass

    @abstractmethod
    def create_type_field_resolver(self, prop: dict, resolve: Callable) -> object:
        pass

    @abstractmethod
    def raise_exception(self, e: Exception):
        pass

    @abstractmethod
    def create_mutation_field_resolver(self, resolver: Callable) -> object:
        pass

    @abstractmethod
    def create_subscription_field_resolver(self, resolver: Callable) -> object:
        pass

    def add_query_property(self, property_name: str, resolver: Callable):
        self._query_properties.append(
            (property_name, self.create_query_field_resolver(resolver))
        )

    def add_mutation_property(self, property_name: str, resolver: Callable):
        self._mutation_properties.append(
            (property_name, self.create_mutation_field_resolver(resolver))
        )

    def add_subscription_property(self, property_name: str, resolver: Callable):
        self._subscription_properties.append(
            (property_name, self.create_subscription_field_resolver(resolver))
        )

    def create_query(self, name: str = "Query") -> Union[Type, None]:
        if len(self._query_properties) == 0:
            return None
        return type(name, (), {k: r for (k, r) in self._query_properties})

    def create_mutation(self, name: str = "Mutation") -> Union[Type, None]:
        if len(self._mutation_properties) == 0:
            return None
        return type(name, (), {k: r for (k, r) in self._mutation_properties})

    def create_subscription(self, name: str = "Subscription") -> Union[Type, None]:
        if len(self._subscription_properties) == 0:
            return None
        return type(name, (), {k: r for (k, r) in self._subscription_properties})

    @abstractmethod
    def create_schema(self, **kwargs) -> Any:
        pass

    @abstractmethod
    def modify_handler_signature(
        self,
        original_handler: Callable,
        wrapper_handler: Callable,
        default_return_type: Any,
    ) -> Type:
        pass

    @abstractmethod
    def create_graphql_asgi_app(
        self, schema: Any, option: GraphqlOption
    ) -> GraphqlASGI:
        pass



from typing import Type

from nestipy.common import Injectable
from nestipy.common.utils import uniq_list
from nestipy.metadata import Reflect, ModuleMetadata


@Injectable()
class DiscoverService:
    _modules = []
    _providers = []
    _controllers = []

    def get_all_controller(self) -> list[Type]:
        return uniq_list(self._controllers)

    def get_all_module(self) -> list:
        return uniq_list(self._modules)

    def get_all_provider(self) -> list[Type]:
        return uniq_list(self._providers)

    def add_controller(self, *ctrl: object):
        self._controllers += list(ctrl)

    def add_provider(self, *services: object):
        self._providers += list(services)

    def add_module(self, *module_ref: object):
        self._modules += list(module_ref)

    def get_module_providers(self, module: Type):
        return [
            p
            for p in self._providers
            if p.__class__ in Reflect.get_metadata(module, ModuleMetadata.Providers, [])
        ]

    def get_module_controllers(self, module: Type):
        return [
            c
            for c in self._controllers
            if c.__class__
            in Reflect.get_metadata(module, ModuleMetadata.Controllers, [])
        ]



class DataRequest:
    def __init__(self, topic: str, data: str):
        pass

class SubRequest:
    def __init__(self, topic: str):
        pass

class UnsubRequest:
    def __init__(self, topic: str):
        pass

class DataResponse:
    def __init__(self, topic: str, data: str):
        pass

class Empty:
    pass



from typing import Callable, Union, Type, Any

from .context_container import RequestContextContainer

TypeAnnotatedCallable = Callable[
    [str, Union[str, None], Type, RequestContextContainer], Any
]


class ParamAnnotation:
    def __init__(
        self,
        callback: TypeAnnotatedCallable,
        key: str,
        token: Union[str | Any, None] = None,
    ):
        self.key: Union[str, None] = key
        self.token: Union[str | Any, None] = token
        self.callback = callback



from .decorator import Command
from .commander import NestipyCommander
from .abstract import BaseCommand
from .factory import CommandFactory

__all__ = ["Command", "NestipyCommander", "BaseCommand", "CommandFactory"]



from typing import Type

from nestipy.common.decorator import Injectable
from nestipy.metadata import SetMetadata

GATEWAY_KEY = "__GATEWAY__"
EVENT_KEY = "__GATEWAY_EVENT__"
SUCCESS_EVENT_KEY = "__GATEWAY_SUCCESS_EVENT__"
ERROR_EVENT_KEY = "__GATEWAY_ERROR_EVENT__"


def Gateway(namespace: str = "/"):
    meta_decorator = SetMetadata(GATEWAY_KEY, namespace)

    def decorator(cls: Type):
        cls = meta_decorator(cls)
        return Injectable()(cls)

    return decorator


def SubscribeMessage(event: str):
    return SetMetadata(EVENT_KEY, event)


def SuccessEvent(event: str):
    return SetMetadata(SUCCESS_EVENT_KEY, event)


def ErrorEvent(event: str):
    return SetMetadata(ERROR_EVENT_KEY, event)



def uniq(data: list) -> list:
    return list(set(data))


def uniq_list(data: list) -> list:
    """
    Make a list to unique value
    :return:
    :rtype:
    :param data:
    :type data:
    :return: []
    :rtype: []
    """
    uniq_data = []
    for d in data:
        if d not in uniq_data:
            uniq_data.append(d)
    return uniq_data


def deep_merge(dict_from: dict, dict_to: dict) -> dict:
    """
    Returns a dict merged value of dict_from and dict_to
        Parameters:
            dict_from (dict): A dictionary value
            dict_to (dict): Another dictionary value
        Returns:
            dict (dict): Merged dict
    """
    keys1 = dict_to.keys()
    keys2 = dict_from.keys()
    for k in keys2:
        if k in keys1:
            dict_to_1 = dict_to[k]
            dict_from_1 = dict_from[k]
            if isinstance(dict_to_1, dict) and isinstance(dict_from_1, dict):
                dict_to_1 = deep_merge(dict_from_1, dict_to_1)
                dict_to[k] = dict_to_1
            elif isinstance(dict_to_1, list) and isinstance(
                dict_from_1,
                list,
            ):
                dict_to[k] = dict_to_1 + dict_from_1
            elif isinstance(
                dict_to_1,
                set,
            ) and isinstance(
                dict_from_1,
                set,
            ):
                dict_to[k] = list(dict_to_1) + list(dict_from_1)
            else:
                pass
        else:
            dict_to[k] = dict_from[k]
    return dict_to


def snakecase_to_camelcase(word: str) -> str:
    """
    Snake case to camelcase
    Args:
        word (str): a snake case string
    Returns:
         word(str): a camelcase string of word
    """
    return " ".join(word.split("_")).capitalize()


if __name__ == "__main__":
    d1: dict = {
        "/test": {"get": {"response": {}, "security": {}, "tags": [], "parameters": {}}}
    }

    d2: dict = {
        "/test": {
            "post": {"response": {}, "security": {}, "tags": [], "parameters": {}}
        }
    }
    res = deep_merge(d1, d2)

    print(res)



class Texts:
    """
    Texts used to generate documentation from OpenAPI Documentation.
    """

    contact: str
    content: str
    email: str
    content_type: str
    example: str
    links: str
    status_code: str
    name: str
    description: str
    parameter: str
    parameters: str
    parameter_location: str
    responses: str
    query: str
    type: str
    default: str
    nullable: str
    request_body: str
    no_parameters: str
    schema: str
    schema_of_the_request_body: str
    schema_of_the_response_body: str
    schemas: str
    about_schemas: str
    details: str
    external_docs: str
    required: str
    properties: str
    yes: str
    no: str
    other_accepted_types: str
    other_possible_types: str
    other_responses: str
    auto_generated_example_note: str
    license: str
    terms_of_service: str
    servers: str
    url: str
    common_responses: str
    common_responses_about: str
    security_schemes: str
    common_parameters_about: str
    scheme: str
    response_headers: str
    for_more_information: str
    tags: str

    def get_yes_no(self, value: bool) -> str:
        return self.yes if value else self.no


class EnglishTexts(Texts):
    contact = "Contact"
    license = "License"
    content = "Content"
    content_type = "Content type"
    email = "Email"
    example = "Example"
    status_code = "Status code"
    name = "Name"
    description = "Description"
    links = "Links"
    parameter = "Parameter"
    parameters = "Input parameters"
    parameter_location = "In"
    responses = "Responses"
    query = "Query"
    type = "Type"
    default = "Default"
    nullable = "Nullable"
    request_body = "Request body"
    no_parameters = "No parameters"
    schema = "Schema"
    schema_of_the_request_body = "Schema of the request body"
    schema_of_the_response_body = "Schema of the response body"
    schemas = "Schemas"
    about_schemas = "This section describes all types of objects handled by the API."
    details = "Details"
    required = "Required"
    properties = "Properties"
    yes = "Yes"
    no = "No"
    other_accepted_types = "Other accepted types"
    other_possible_types = "Other possible types"
    other_responses = "Other responses"
    auto_generated_example_note = (
        "This example has been generated automatically from the schema and it is "
        "not accurate. Refer to the schema for more information."
    )
    terms_of_service = "Terms of service"
    servers = "Servers"
    url = "URL"
    common_responses = "Common responses"
    common_responses_about = (
        "This section describes common responses that are reused across operations."
    )
    common_parameters = "Common parameters"
    common_parameters_about = (
        "This section describes common parameters that are reused across operations."
    )
    security_schemes = "Security schemes"
    scheme = "Scheme"
    response_headers = "Response headers"
    external_docs = "More documentation"
    for_more_information = "For more information"
    tags = "Tags"



from typing import Callable

import uvicorn

from request import Request
from response import Response


async def app(scope: dict, receive: Callable, send: Callable):
    if scope.get("type") == "http":
        request = Request(scope, receive, send)
        json = await request.json()
        print(request.query_params, json)
        res = Response()
        await res.status(200).json({"test": "test"})


if __name__ == "__main__":
    uvicorn.run("test:app", reload=True)



import inspect
from typing import Annotated, Union, Callable

from nestipy.common import Module
from nestipy.core import DiscoverService
from nestipy.dynamic_module import NestipyModule
from nestipy.ioc import Inject, ModuleProviderDict
from nestipy.metadata import Reflect
from .event_builder import ConfigurableClassBuilder, EventOption
from .event_emitter import EventEmitter
from .event_metadata import EventMetadata, EventData


@Module(
    providers=[ModuleProviderDict(token=EventEmitter, value=EventEmitter())],
    exports=[EventEmitter],
)
class EventEmitterModule(ConfigurableClassBuilder, NestipyModule):
    _discovery: Annotated[DiscoverService, Inject()]
    _event_emitter: Annotated[EventEmitter, Inject()]

    @classmethod
    def _is_listener(cls, method: callable):
        v = Reflect.get_metadata(method, EventMetadata.Event)
        return v is not None

    def _register_listener(self, method: callable):
        event_data: Union[EventData, None] = Reflect.get_metadata(
            method, EventMetadata.Event, None
        )
        if event_data:
            callback = self._create_async_handler(method)
            if event_data.once:
                self._event_emitter.once(event_data.name, callback)
            else:
                self._event_emitter.add_listener(event_data.name, callback)

    @classmethod
    def _create_async_handler(cls, callback: Callable) -> callable:
        async def async_handler(*args, **kwargs):
            if inspect.iscoroutinefunction(callback):
                return await callback(*args, **kwargs)
            else:
                return callback(*args, **kwargs)

        return async_handler

    def _register_listeners(self):
        instances = (
            self._discovery.get_all_controller() + self._discovery.get_all_provider()
        )
        for p in instances:
            elements = inspect.getmembers(
                p,
                lambda a: inspect.isfunction(a)
                or inspect.iscoroutinefunction(a)
                or inspect.ismethod(a),
            )
            methods = [
                method
                for (method, _) in elements
                if not method.startswith("__") and self._is_listener(getattr(p, method))
            ]
            for m in methods:
                self._register_listener(getattr(p, m))

    async def on_startup(self):
        self._register_listeners()

    async def on_shutdown(self):
        pass

    @classmethod
    def for_root(cls, is_global: bool = True):
        return cls.register(EventOption(), extras={"is_global": is_global})



from typing import Union, Any, Callable

from nestipy.common.decorator import Injectable
from nestipy.common.exception.http import HttpException
from nestipy.common.exception.interface import ExceptionFilter
from nestipy.common.exception.meta import ExceptionKey
from nestipy.common.helpers import SpecialProviderExtractor
from nestipy.common.utils import uniq_list
from nestipy.core.constant import APP_FILTER
from nestipy.core.context.argument_host import ArgumentHost
from nestipy.ioc import NestipyContainer
from nestipy.metadata import Reflect, ClassMetadata


@Injectable()
class ExceptionFilterHandler(SpecialProviderExtractor):
    context: ArgumentHost = None

    def __init__(
        self,
    ):
        self.container = NestipyContainer.get_instance()

    async def catch(
        self, exception: HttpException, context: ArgumentHost, is_http: bool = True
    ) -> Union[Any, None]:
        self.context = context
        handler_module_class = self.context.get_module()
        handler_class = self.context.get_class()
        handler = self.context.get_handler()
        global_filters = context.get_adapter().get_global_filters() if is_http else []
        module_filters = self.extract_special_providers(
            handler_module_class, ExceptionFilter, APP_FILTER
        )
        class_filters = Reflect.get_metadata(handler_class, ExceptionKey.MetaFilter, [])
        handler_filters = Reflect.get_metadata(handler, ExceptionKey.MetaFilter, [])
        all_filters = uniq_list(
            global_filters + module_filters + class_filters + handler_filters
        )
        # setup dependency as the same as the container
        for fit in all_filters:
            if issubclass(fit, ExceptionFilter):
                # Put dependency
                services = self.container.get_all_services()
                Reflect.set_metadata(
                    fit,
                    ClassMetadata.Metadata,
                    ClassMetadata(handler_class, global_providers=services),
                )
        return await self._apply_exception_filter(exception, 0, all_filters, None)

    async def _apply_exception_filter(
        self, exception: "HttpException", index: int, all_filters: list, result: Any
    ):
        if len(all_filters) > index:
            exception_filter = all_filters[index]
            all_exception_to_catch = Reflect.get_metadata(
                exception_filter, ExceptionKey.MetaType, []
            )
            new_result = await self._apply_exception_filter_for_exception(
                exception_filter, exception, 0, all_exception_to_catch, result
            )
            return await self._apply_exception_filter(
                exception,
                index + 1,
                all_filters,
                new_result if new_result is not None else result,
            )
        else:
            return result

    async def _apply_exception_filter_for_exception(
        self,
        exception_filter: Union[Callable, "ExceptionFilter"],
        exception: "HttpException",
        index: int,
        all_exception_to_catch: list,
        result: Any,
    ):
        if len(all_exception_to_catch) > index:
            exception_to_catch = all_exception_to_catch[index]
            if isinstance(exception, exception_to_catch) or (
                exception.status_code == exception_to_catch.status_code
            ):
                new_result = await self._catch_exception(exception_filter, exception)
                return await self._apply_exception_filter_for_exception(
                    exception_filter,
                    exception,
                    index + 1,
                    all_exception_to_catch,
                    new_result if new_result is not None else result,
                )
            return result
        else:
            return await self._catch_exception(exception_filter, exception)

    async def _catch_exception(
        self,
        exception_filter: Union[Callable, "ExceptionFilter"],
        exception: "HttpException",
    ):
        instance = (
            await self.container.get(exception_filter)
            if not isinstance(exception_filter, ExceptionFilter)
            else exception_filter
        )
        return await instance.catch(exception, self.context)



from __future__ import annotations

import re
from urllib.parse import parse_qsl as _parse_qsl

__all__ = ("parse_content_header",)

_token = r"([\w!#$%&'*+\-.^_`|~]+)"  # noqa: S105
_quoted = r'"([^"]*)"'
_param = re.compile(rf";\s*{_token}=(?:{_token}|{_quoted})", re.ASCII)
_firefox_quote_escape = re.compile(r'\\"(?!; |\s*$)')


def parse_qsl(qs: bytes, separator: str) -> list[tuple[str, str]]:
    return _parse_qsl(qs.decode("utf-8"), keep_blank_values=True, separator=separator)


def parse_content_header(value: str) -> tuple[str, dict[str, str]]:
    value = _firefox_quote_escape.sub("%22", value)
    pos = value.find(";")
    if pos == -1:
        options: dict[str, str] = {}
    else:
        options = {
            m.group(1).lower(): m.group(2) or m.group(3).replace("%22", '"')
            for m in _param.finditer(value[pos:])
        }
        value = value[:pos]
    return value.strip().lower(), options



import inspect
from typing import Type, Union

from nestipy.common.decorator import Injectable
from nestipy.common.guards.can_activate import CanActivate
from nestipy.common.guards.meta import GuardKey
from nestipy.common.helpers import SpecialProviderExtractor
from nestipy.core.constant import APP_GUARD
from nestipy.core.context.execution_context import ExecutionContext
from nestipy.ioc import NestipyContainer
from nestipy.metadata import ClassMetadata, Reflect


@Injectable()
class GuardProcessor(SpecialProviderExtractor):
    def __init__(self):
        self.container = NestipyContainer.get_instance()

    async def process(
        self, context: ExecutionContext, is_http: bool = True
    ) -> tuple[bool, Union[Type, None]]:
        handler_module_class = context.get_module()
        handler_class = context.get_class()
        handler = context.get_handler()
        global_guards = []
        if is_http:
            global_guards = context.get_adapter().get_global_guards()
        module_guards = self.extract_special_providers(
            handler_module_class, CanActivate, APP_GUARD
        )
        handler_class_guards = Reflect.get_metadata(handler_class, GuardKey.Meta, [])
        handler_guards = Reflect.get_metadata(handler, GuardKey.Meta, [])

        for g in handler_guards + handler_class_guards + module_guards + global_guards:
            if inspect.isclass(g) and issubclass(g, CanActivate):
                services = self.container.get_all_services()
                # Put dependency
                Reflect.set_metadata(
                    g,
                    ClassMetadata.Metadata,
                    ClassMetadata(handler_class, global_providers=services),
                )
                #  load instance from container
                instance: CanActivate = await self.container.get(g)
                callback = instance.can_activate
                # check if can_activate is coroutine
                if inspect.iscoroutinefunction(callback):
                    can_activate: bool = await callback(context)
                else:
                    can_activate: bool = callback(context)
                if not can_activate:
                    return False, g
        return True, None



from typing import Callable

from nestipy.common.decorator import Injectable
from nestipy.metadata.decorator import SetMetadata
from .meta import CommanderMeta


class Command:
    def __init__(self, name: str, desc: str = ""):
        self.name = name
        self.desc = desc

    def __call__(self, func: Callable) -> Callable:
        data = {"name": self.name, "description": self.desc}
        decorator = SetMetadata(CommanderMeta.Meta, data)
        return decorator(Injectable()(func))



import asyncio
import uuid
from abc import ABC, abstractmethod
from dataclasses import asdict
from dataclasses import field, dataclass
from typing import AsyncIterator, Any, Union
from typing import Optional

from nestipy.microservice.context import RpcRequest, RpcResponse, MICROSERVICE_CHANNEL
from nestipy.microservice.exception import RpcException, RPCErrorMessage, RPCErrorCode
from nestipy.microservice.serializer import Serializer, JSONSerializer
from .option import RabbitMQClientOption, TCPClientOption, GrpcClientOption
from .option import RedisClientOption, MqttClientOption, NatsClientOption
from .option import Transport


@dataclass
class MicroserviceOption:
    transport: Transport = field(default=Transport.TCP)
    option: Optional[
        Union[
            RedisClientOption,
            MqttClientOption,
            NatsClientOption,
            RabbitMQClientOption,
            TCPClientOption,
            GrpcClientOption,
        ]
    ] = None
    proxy: Optional["ClientProxy"] = None
    serializer: Serializer = field(default=JSONSerializer())
    channel_key: str = field(default="nestipy")
    timeout: int = 30

    def __post_init__(self):
        if (
            self.transport == Transport.TCP
            and self.option is not None
            and not isinstance(self.option, TCPClientOption)
        ):
            raise ValueError("Option must be of type TCPClientOption for TCP transport")
        elif (
            self.transport == Transport.REDIS
            and self.option is not None
            and not isinstance(self.option, RedisClientOption)
        ):
            raise ValueError(
                "Option must be of type RedisClientOption for REDIS transport"
            )
        elif (
            self.transport == Transport.MQTT
            and self.option is not None
            and not isinstance(self.option, MqttClientOption)
        ):
            raise ValueError(
                "Option must be of type MqttClientOption for MQTT transport"
            )
        elif (
            self.transport == Transport.NATS
            and self.option is not None
            and not isinstance(self.option, NatsClientOption)
        ):
            raise ValueError(
                "Option must be of type NatsClientOptions for NATS transport"
            )
        elif (
            self.transport == Transport.RABBITMQ
            and self.option is not None
            and not isinstance(self.option, RabbitMQClientOption)
        ):
            raise ValueError(
                "Option must be of type RabbitMQClientOption for RABBITMQ transport"
            )

        elif (
            self.transport == Transport.GRPC
            and self.option is not None
            and not isinstance(self.option, GrpcClientOption)
        ):
            raise ValueError(
                "Option must be of type GrpcClientOption for GRPC transport"
            )


class ClientProxy(ABC):
    def __init__(self, option: MicroserviceOption):
        self.option = option

    @abstractmethod
    async def slave(self) -> "ClientProxy":
        pass

    @abstractmethod
    async def connect(self):
        pass

    @abstractmethod
    async def _publish(self, topic, data):
        pass

    @abstractmethod
    async def subscribe(self, *args, **kwargs):
        pass

    @abstractmethod
    async def unsubscribe(self, *args):
        pass

    @abstractmethod
    async def send_response(self, topic, data):
        pass

    async def before_start(self):
        await asyncio.sleep(0.001)

    async def before_close(self):
        await asyncio.sleep(0.001)

    async def send(
        self, topic, data: Any, headers: dict[str, str] = None
    ) -> RpcResponse:
        await self.connect()
        response_topic = uuid.uuid4().hex
        request = RpcRequest(
            pattern=topic,
            data=data,
            response_topic=response_topic,
            headers=headers or {},
        )
        response_channel = f"{MICROSERVICE_CHANNEL}:response:{request.response_topic}"
        await self.subscribe(response_channel)
        await self._publish(
            f"{MICROSERVICE_CHANNEL}:{self.option.channel_key}",
            await self.option.serializer.serialize(asdict(request)),
        )
        #  add timeout
        try:
            async with asyncio.timeout(self.option.timeout):
                rpc_response = await self.listen_response(
                    response_channel, self.option.timeout
                )
        except asyncio.TimeoutError:
            raise RpcException(
                status_code=RPCErrorCode.DEADLINE_EXCEEDED,
                message=RPCErrorMessage.DEADLINE_EXCEEDED,
            )
        await self.unsubscribe(response_channel)
        await self.close()
        json_rep = await self.option.serializer.deserialize(rpc_response)
        response = RpcResponse.from_dict(json_rep)
        if response.exception:
            raise response.exception
        return response

    async def emit(self, topic, data, headers: dict[str, str] = None):
        request = RpcRequest(pattern=topic, data=data, headers=headers or {})
        json_req = await self.option.serializer.serialize(asdict(request))
        await self._publish(
            f"{MICROSERVICE_CHANNEL}:{self.option.channel_key}", json_req
        )

    @abstractmethod
    def listen(self) -> AsyncIterator[str]:
        pass

    @abstractmethod
    async def listen_response(self, from_topic: str, timeout: int = 30) -> str:
        pass

    @abstractmethod
    async def close(self):
        pass



from nestipy.core.providers.async_local_storage import AsyncLocalStorage
from nestipy.core.providers.background import BackgroundTask, BackgroundTasks
from nestipy.core.providers.discover import DiscoverService
from .adapter.http_adapter import HttpAdapter
from .constant import AppKey
from .context.execution_context import ExecutionContext, ArgumentHost, HttpArgumentHost
from .middleware import MiddlewareConsumer
from .nestipy_application import NestipyApplication, NestipyConfig
from .nestipy_factory import NestipyFactory
from .on_destroy import OnDestroy
from .on_init import OnInit
from .platform import FastApiApplication, BlackSheepApplication
from .template import MinimalJinjaTemplateEngine

try:
    from .nestipy_microservice import NestipyMicroservice, NestipyConnectMicroservice
except ImportError:
    NestipyMicroservice = None
    NestipyConnectMicroservice = None

__all__ = [
    "NestipyFactory",
    "NestipyApplication",
    "NestipyConfig",
    "HttpAdapter",
    "ExecutionContext",
    "ArgumentHost",
    "HttpArgumentHost",
    "FastApiApplication",
    "BlackSheepApplication",
    "MinimalJinjaTemplateEngine",
    "AppKey",
    "MiddlewareConsumer",
    "DiscoverService",
    "AsyncLocalStorage",
    "OnInit",
    "OnDestroy",
    "NestipyMicroservice",
    "NestipyConnectMicroservice",
    "BackgroundTask",
    "BackgroundTasks",
]
